"""
Supporting classes for COAR Notify server implementations
"""
import json
import typing
from typing import Union

from coarnotify.factory import COARNotifyFactory

if typing.TYPE_CHECKING:
    from coarnotify.core.notify import NotifyPattern


class COARNotifyReceipt:
    """
    An object representing the response from a COAR Notify server.

    Server implementations should construct and return this object with the appropriate properties
    when implementing the :py:meth:`COARNotifyServiceBinding.notification_received` binding

    :param status: the HTTP status code, should be one of the constants ``CREATED`` (201) or ``ACCEPTED`` (202)
    :param location: the HTTP URI for the resource that was created (if present)
    """

    CREATED = 201
    """The status code for a created resource"""

    ACCEPTED = 202
    """The status code for an accepted request"""

    def __init__(self, status: int, location: str = None):
        """
        Construct a new COARNotifyReceipt object with the status code and location URL (optional)

        :param status: the HTTP status code, should be one of the constants ``CREATED`` (201) or ``ACCEPTED`` (202)
        :param location: the HTTP URI for the resource that was created (if present)
        """
        self._status = status
        self._location = location

    @property
    def status(self) -> int:
        """The status code of the response.  Should be one of the constants ``CREATED`` (201) or ``ACCEPTED`` (202)"""
        return self._status

    @property
    def location(self) -> Union[str, None]:
        """The HTTP URI of the created resource, if present"""
        return self._location


class COARNotifyServiceBinding:
    """
    Interface for implementing a COAR Notify server binding.

    Server implementation should extend this class and implement the :py:meth:`notification_received` method

    That method will receive a :py:class:`NotifyPattern` object, which will be one of the known types
    and should return a :py:class:`COARNotifyReceipt` object with the appropriate status code and location URL
    """

    def notification_received(self, notification: 'NotifyPattern') -> COARNotifyReceipt:
        """
        Process the receipt of the given notification, and respond with an appropriate receipt object

        :param notification: the notification object received
        :return: the receipt object to send back to the client
        """
        raise NotImplementedError()


class COARNotifyServerError(Exception):
    """
    An exception class for server errors in the COAR Notify server implementation.

    The web layer of your server implementation should be able to intercept this from the
    :py:meth:`COARNotifyServer.receive` method and return the appropriate HTTP status code and message to the
    user in its standard way.

    :param status: HTTP Status code to respond to the client with
    :param msg: Message to send back to the client
    """

    def __init__(self, status: int, msg: str):
        """
        Construct a new COARNotifyServerError with the given status code and message

        :param status: HTTP Status code to respond to the client with
        :param msg: Message to send back to the client
        """
        self._status = status
        self._msg = msg
        super(COARNotifyServerError, self).__init__(msg)

    @property
    def status(self) -> int:
        """HTTP status code for the error"""
        return self._status

    @property
    def message(self) -> str:
        """The error message"""
        return self._msg


class COARNotifyServer:
    """
    The main entrypoint to the COAR Notify server implementation.

    The web layer of your application should pass the json/raw payload of any incoming notification to the
    :py:meth:`receive` method, which will parse the payload and pass it to the :py:meth:`COARNotifyServiceBinding.notification_received`
    method of your service implementation

    This object should be constructed with your service implementation passed to it, for example

    .. code-block:: python

        server = COARNotifyServer(MyServiceBinding())
        try:
            response = server.receive(request.json)
            return jsonify(response)
        except COARNotifyServerError as e:
            abort(e.status, e.message)

    :param service_impl: Your service implementation
    """

    def __init__(self, service_impl: COARNotifyServiceBinding):
        """
        Construct a new COARNotifyServer with the given service implementation
        :param service_impl: Your service implementation
        """
        self._service_impl = service_impl

    def receive(self, raw: Union[dict, str], validate: bool = True) -> COARNotifyReceipt:
        """
        Receive an incoming notification as JSON, parse and validate (optional) and then pass to the
        service implementation

        :param raw: The JSON representation of the data, either as a string or a dictionary
        :param validate:    Whether to validate the notification before passing to the service implementation
        :return:   The COARNotifyReceipt response from the service implementation
        """
        if isinstance(raw, str):
            raw = json.loads(raw)

        obj = COARNotifyFactory.get_by_object(raw)
        if validate:
            if not obj.validate():
                raise COARNotifyServerError(400, "Invalid notification")

        return self._service_impl.notification_received(obj)



"""
Pattern to represent a Request Review notification
https://coar-notify.net/specification/1.0.0/request-review/
"""
from coarnotify.core.notify import NotifyPattern, NotifyTypes, NotifyObject, NotifyItem, NotifyProperties
from coarnotify.core.activitystreams2 import ActivityStreamsTypes, Properties
from coarnotify.exceptions import ValidationError

from typing import Union

__all__ = ["RequestReview", "RequestReviewObject", "RequestReviewItem"]

class RequestReview(NotifyPattern):
    """
    Class to represent a Request Review notification
    """

    TYPE = [ActivityStreamsTypes.OFFER, NotifyTypes.REVIEW_ACTION]
    """Request Review types, including an ActivityStreams offer and a COAR Notify Review Action"""

    @property
    def object(self) -> Union["RequestReviewObject", None]:
        """
        Custom getter to retrieve the object property as a RequestReviewObject
        :return:
        """
        o = self.get_property(Properties.OBJECT)
        if o is not None:
            return RequestReviewObject(o,
                                    validate_stream_on_construct=False,
                                    validate_properties=self.validate_properties,
                                    validators=self.validators,
                                    validation_context=Properties.OBJECT,
                                    properties_by_reference=self._properties_by_reference)
        return None


class RequestReviewObject(NotifyObject):
    """
    Custom Request Review Object class to return the custom RequestReviewItem class
    """
    @property
    def item(self) -> Union["RequestReviewItem", None]:
        """
        Custom getter to retrieve the item property as a RequestReviewItem
        :return:
        """
        i = self.get_property(NotifyProperties.ITEM)
        if i is not None:
            return RequestReviewItem(i,
                              validate_stream_on_construct=False,
                              validate_properties=self.validate_properties,
                              validators=self.validators,
                              validation_context=NotifyProperties.ITEM,
                              properties_by_reference=self._properties_by_reference)
        return None


class RequestReviewItem(NotifyItem):
    """
    Custom Request Review Item class to provide the custom validation
    """

    def validate(self) -> bool:
        """
        Extend the base validation to include the following constraints:

        * The type property is required and must validate
        * the ``mediaType`` property is required

        :return: ``True`` if validation passes, else raise a ``ValidationError``
        """
        ve = ValidationError()
        try:
            super(RequestReviewItem, self).validate()
        except ValidationError as superve:
            ve = superve

        self.required_and_validate(ve, Properties.TYPE, self.type)
        self.required(ve, NotifyProperties.MEDIA_TYPE, self.media_type)

        if ve.has_errors():
            raise ve
        return True


"""
Pattern to represent a Request Endorsement notification
https://coar-notify.net/specification/1.0.0/request-endorsement/
"""
from coarnotify.core.notify import NotifyPattern, NotifyTypes, NotifyItem, NotifyProperties, NotifyObject
from coarnotify.core.activitystreams2 import ActivityStreamsTypes, Properties
from coarnotify.exceptions import ValidationError

from typing import Union

__all__ = ["RequestEndorsement", "RequestEndorsementObject", "RequestEndorsementItem"]

class RequestEndorsement(NotifyPattern):
    """
    Class to represent a Request Endorsement notification
    """
    TYPE = [ActivityStreamsTypes.OFFER, NotifyTypes.ENDORSMENT_ACTION]
    """Request Endorsement types, including an ActivityStreams offer and a COAR Notify Endorsement Action"""

    @property
    def object(self) -> Union["RequestEndorsementObject", None]:
        """
        Custom getter to retrieve the object property as a RequestEndorsementObject

        :return:
        """
        o = self.get_property(Properties.OBJECT)
        if o is not None:
            return RequestEndorsementObject(o,
                                        validate_stream_on_construct=False,
                                        validate_properties=self.validate_properties,
                                        validators=self.validators,
                                        validation_context=Properties.OBJECT,
                                        properties_by_reference=self._properties_by_reference)
        return None


class RequestEndorsementObject(NotifyObject):
    """
    Custom object class for Request Endorsement to provide the custom item getter
    """

    @property
    def item(self) -> Union["RequestEndorsementItem", None]:
        """
        Custom getter to retrieve the item property as a RequestEndorsementItem
        :return:
        """
        i = self.get_property(NotifyProperties.ITEM)
        if i is not None:
            return RequestEndorsementItem(i,
                              validate_stream_on_construct=False,
                              validate_properties=self.validate_properties,
                              validators=self.validators,
                              validation_context=NotifyProperties.ITEM,
                              properties_by_reference=self._properties_by_reference)
        return None


class RequestEndorsementItem(NotifyItem):
    """
    Custom item class for Request Endorsement to provide the custom validation
    """

    def validate(self) -> bool:
        """
        Extend the base validation to include the following constraints:

        * The item type is required and must validate
        * The ``mediaType`` property is required

        :return: ``True`` if validation passes, otherwise raise a ``ValidationError``
        """
        ve = ValidationError()
        try:
            super(RequestEndorsementItem, self).validate()
        except ValidationError as superve:
            ve = superve

        self.required_and_validate(ve, Properties.TYPE, self.type)
        self.required(ve, NotifyProperties.MEDIA_TYPE, self.media_type)

        if ve.has_errors():
            raise ve
        return True



from copy import deepcopy

from coarnotify.test.fixtures import BaseFixtureFactory


class UndoOfferFixtureFactory(BaseFixtureFactory):
    @classmethod
    def source(cls, copy=True):
        if copy:
            return deepcopy(UNDO_OFFER)
        return UNDO_OFFER

    @classmethod
    def invalid(cls):
        source = cls.source()
        cls._base_invalid(source)
        cls._actor_invalid(source)
        return source

UNDO_OFFER = {
    "@context": [
        "https://www.w3.org/ns/activitystreams",
        "https://coar-notify.net"
    ],
    "actor": {
        "id": "https://some-organisation.org",
        "name": "Some Organisation",
        "type": "Organization"
    },
    "id": "urn:uuid:46956915-e3fe-4528-8789-1d325a356e4f",
    "inReplyTo": "urn:uuid:0370c0fb-bb78-4a9b-87f5-bed307a509dd",
    "object": {
        "actor": {
            "id": "https://orcid.org/0000-0002-1825-0097",
            "name": "Josiah Carberry",
            "type": "Person"
        },
        "id": "urn:uuid:0370c0fb-bb78-4a9b-87f5-bed307a509dd",
        "object": {
            "id": "https://research-organisation.org/repository/preprint/201203/421/",
            "ietf:cite-as": "https://doi.org/10.5555/12345680",
            "ietf:item": {
                "id": "https://research-organisation.org/repository/preprint/201203/421/content.pdf",
                "mediaType": "application/pdf",
                "type": [
                    "Page",
                    "sorg:AboutPage"
                ]
            },
            "type": "sorg:AboutPage"
        },
        "origin": {
            "id": "https://research-organisation.org/repository",
            "inbox": "https://research-organisation.org/inbox/",
            "type": "Service"
        },
        "target": {
            "id": "https://overlay-journal.com/system",
            "inbox": "https://overlay-journal.com/inbox/",
            "type": "Service"
        },
        "type": [
            "Offer",
            "coar-notify:EndorsementAction"
        ]
    },
    "origin": {
        "id": "https://some-organisation.org",
        "inbox": "https://some-organisation.org/inbox/",
        "type": "Service"
    },
    "summary": "The offer has been withdrawn because...",
    "target": {
        "id": "https://generic-service.com/system",
        "inbox": "https://generic-service.com/system/inbox/",
        "type": "Service"
    },
    "type": "Undo"
}


class BaseFixtureFactory:
    @classmethod
    def source(cls, copy=True):
        raise NotImplementedError()

    @classmethod
    def invalid(cls):
        source = cls.source()
        cls._base_invalid(source)
        return source

    @classmethod
    def expected_value(cls, path):
        source = cls.source(copy=False) # we're only reading the value, so no need to clone it
        return cls._value_from_dict(path, source)

    @classmethod
    def _base_invalid(cls, source):
        source["id"] = "not a uri"
        source["inReplyTo"] = "not a uri"
        source["origin"]["id"] = "urn:uuid:4fb3af44-d4f8-4226-9475-2d09c2d8d9e0"
        source["origin"]["inbox"] = "not a uri"
        source["origin"]["type"] = "NotAValidType"
        source["target"]["id"] = "urn:uuid:4fb3af44-d4f8-4226-9475-2d09c2d8d9e0"
        source["target"]["inbox"] = "not a uri"
        source["target"]["type"] = "NotAValidType"
        source["type"] = "NotAValidType"
        return source

    @classmethod
    def _actor_invalid(self, source):
        source["actor"]["id"] = "not a uri"
        source["actor"]["type"] = "NotAValidType"
        return source

    @classmethod
    def _object_invalid(self, source):
        source["object"]["id"] = "not a uri"
        source["object"]["cite_as"] = "urn:uuid:4fb3af44-d4f8-4226-9475-2d09c2d8d9e0"
        return source

    @classmethod
    def _context_invalid(self, source):
        source["context"]["id"] = "not a uri"
        source["context"]["type"] = "NotAValidType"
        source["context"]["cite_as"] = "urn:uuid:4fb3af44-d4f8-4226-9475-2d09c2d8d9e0"
        return source

    @classmethod
    def _value_from_dict(cls, path, dictionary):
        bits = path.split(".")
        node = dictionary
        for bit in bits:
            node = node[bit]
        return node


from unittest import TestCase

from coarnotify.client import COARNotifyClient
from coarnotify.patterns import (
    Accept,
    AnnounceEndorsement,
    AnnounceRelationship,
    AnnounceReview,
    AnnounceServiceResult,
    Reject,
    RequestEndorsement,
    RequestReview,
    TentativelyAccept,
    TentativelyReject,
    UnprocessableNotification,
    UndoOffer
)

from coarnotify.test.fixtures import (
    AcceptFixtureFactory,
    AnnounceEndorsementFixtureFactory,
    AnnounceRelationshipFixtureFactory,
    AnnounceReviewFixtureFactory,
    AnnounceServiceResultFixtureFactory,
    RejectFixtureFactory,
    RequestEndorsementFixtureFactory,
    RequestReviewFixtureFactory,
    TentativelyAcceptFixtureFactory,
    TentativelyRejectFixtureFactory,
    UnprocessableNotificationFixtureFactory,
    UndoOfferFixtureFactory
)

INBOX = "http://localhost:5005/inbox"


class TestClient(TestCase):
    def test_01_accept(self):
        client = COARNotifyClient(INBOX)
        source = AcceptFixtureFactory.source()
        acc = Accept(source)
        resp = client.send(acc)
        assert resp.action == resp.CREATED
        assert resp.location is not None
        print(resp.location)

    def test_02_announce_endorsement(self):
        client = COARNotifyClient(INBOX)
        source = AnnounceEndorsementFixtureFactory.source()
        ae = AnnounceEndorsement(source)
        resp = client.send(ae)
        assert resp.action == resp.CREATED
        assert resp.location is not None
        print(resp.location)

    def test_04_announce_relationship(self):
        client = COARNotifyClient(INBOX)
        source = AnnounceRelationshipFixtureFactory.source()
        ae = AnnounceRelationship(source)
        resp = client.send(ae)
        assert resp.action == resp.CREATED
        assert resp.location is not None
        print(resp.location)

    def test_05_announce_review(self):
        client = COARNotifyClient(INBOX)
        source = AnnounceReviewFixtureFactory.source()
        ae = AnnounceReview(source)
        resp = client.send(ae)
        assert resp.action == resp.CREATED
        assert resp.location is not None
        print(resp.location)

    def test_06_announce_service_result(self):
        client = COARNotifyClient(INBOX)
        source = AnnounceServiceResultFixtureFactory.source()
        ae = AnnounceServiceResult(source)
        resp = client.send(ae)
        assert resp.action == resp.CREATED
        assert resp.location is not None
        print(resp.location)

    def test_07_reject(self):
        client = COARNotifyClient(INBOX)
        source = RejectFixtureFactory.source()
        ae = Reject(source)
        resp = client.send(ae)
        assert resp.action == resp.CREATED
        assert resp.location is not None
        print(resp.location)

    def test_08_request_endorsement(self):
        client = COARNotifyClient(INBOX)
        source = RequestEndorsementFixtureFactory.source()
        ae = RequestEndorsement(source)
        resp = client.send(ae)
        assert resp.action == resp.CREATED
        assert resp.location is not None
        print(resp.location)

    def test_09_request_review(self):
        client = COARNotifyClient(INBOX)
        source = RequestReviewFixtureFactory.source()
        ae = RequestReview(source)
        resp = client.send(ae)
        assert resp.action == resp.CREATED
        assert resp.location is not None
        print(resp.location)

    def test_10_tentatively_accept(self):
        client = COARNotifyClient(INBOX)
        source = TentativelyAcceptFixtureFactory.source()
        ae = TentativelyAccept(source)
        resp = client.send(ae)
        assert resp.action == resp.CREATED
        assert resp.location is not None
        print(resp.location)

    def test_11_tentatively_reject(self):
        client = COARNotifyClient(INBOX)
        source = TentativelyRejectFixtureFactory.source()
        ae = TentativelyReject(source)
        resp = client.send(ae)
        assert resp.action == resp.CREATED
        assert resp.location is not None
        print(resp.location)

    def test_12_unprocessable_notification(self):
        client = COARNotifyClient(INBOX)
        source = UnprocessableNotificationFixtureFactory.source()
        ae = UnprocessableNotification(source)
        resp = client.send(ae)
        assert resp.action == resp.CREATED
        assert resp.location is not None
        print(resp.location)

    def test_13_undo_offer(self):
        client = COARNotifyClient(INBOX)
        source = UndoOfferFixtureFactory.source()
        ae = UndoOffer(source)
        resp = client.send(ae)
        assert resp.action == resp.CREATED
        assert resp.location is not None
        print(resp.location)



from unittest import TestCase
from copy import deepcopy

from coarnotify.exceptions import ValidationError

from coarnotify.core.notify import NotifyPattern, NotifyService, NotifyObject, NotifyActor, NotifyItem
from coarnotify.patterns import (
    Accept,
    AnnounceEndorsement,
    AnnounceRelationship,
    AnnounceReview,
    AnnounceServiceResult,
    Reject,
    RequestEndorsement,
    RequestReview,
    TentativelyAccept,
    TentativelyReject,
    UnprocessableNotification,
    UndoOffer
)
from coarnotify.test.fixtures.notify import NotifyFixtureFactory
from coarnotify.test.fixtures import (
    AcceptFixtureFactory,
    AnnounceEndorsementFixtureFactory,
    AnnounceRelationshipFixtureFactory,
    AnnounceReviewFixtureFactory,
    AnnounceServiceResultFixtureFactory,
    RejectFixtureFactory,
    RequestEndorsementFixtureFactory,
    RequestReviewFixtureFactory,
    TentativelyAcceptFixtureFactory,
    TentativelyRejectFixtureFactory,
    UnprocessableNotificationFixtureFactory,
    UndoOfferFixtureFactory
)


class TestModels(TestCase):

    def _get_testable_properties(self, source, prop_map=None):
        def expand(node, path):
            paths = []
            for k, v in node.items():
                if isinstance(v, dict):
                    paths += expand(v, f"{path}.{k}")
                else:
                    paths.append(f"{path}.{k}")
            paths = [p[1:] if p.startswith(".") else p for p in paths if "@context" not in p]  # strip the leading "."
            return paths

        obj_properties = expand(source, "")

        if prop_map is None:
            prop_map = {
                "inReplyTo": "in_reply_to",
                "context.ietf:cite-as": "context.cite_as",
                "context.ietf:item.id": "context.item.id",
                "context.ietf:item.mediaType": "context.item.media_type",
                "context.ietf:item.type": "context.item.type",
                "object.as:subject": "object.triple[2]",
                "object.as:relationship": "object.triple[1]",
                "object.as:object": "object.triple[0]",
                "object.ietf:cite-as": "object.cite_as",
                "object.ietf:item.id": "object.item.id",
                "object.ietf:item.mediaType": "object.item.media_type",
                "object.ietf:item.type": "object.item.type",
                "object.object.ietf:cite-as": "object.object.cite_as",
                "object.object.ietf:item.id": "object.object.item.id",
                "object.object.ietf:item.mediaType": "object.object.item.media_type",
                "object.object.ietf:item.type": "object.object.item.type",
            }

        proptest = [p if p not in prop_map else (prop_map[p], p) for p in obj_properties]
        return proptest

    def _apply_property_test(self, proptest, obj, fixtures):
        def get_prop(source, prop):
            p = prop
            if isinstance(prop, tuple):
                p = prop[1]
            bits = p.split(".")
            for bit in bits:
                idx = None
                if "[" in bit:
                    bit, idx = bit.split("[")
                    idx = int(idx[:-1])
                source = getattr(source, bit)
                if idx is not None:
                    source = source[idx]
            return source

        for prop in proptest:
            if isinstance(prop, tuple):
                oprop = prop[0]
                fprop = prop[1]
            else:
                oprop = prop
                fprop = prop

            print(oprop, fprop)
            oval = get_prop(obj, oprop)
            eval = fixtures.expected_value(fprop)

            # allow a single value to be equivalent to a list containing a single value
            if isinstance(oval, list) and len(oval) == 1:
                oval = oval[0]
            if isinstance(eval, list) and len(eval) == 1:
                eval = eval[0]

            assert oval == eval, f"{oprop}:{oval} - {fprop}:{eval}"

    def test_01_notify_manual_construct(self):
        n = NotifyPattern()

        # check the default properties
        assert n.id is not None
        assert n.id.startswith("urn:uuid:")
        assert n.type == NotifyPattern.TYPE
        assert n.origin is None
        assert n.target is None
        assert n.object is None
        assert n.actor is None
        assert n.in_reply_to is None
        assert n.context is None

        # now check the setters
        n.id = "urn:whatever"
        n.ALLOWED_TYPES = ["Object", "Other"]   # this is a hack to test the setter
        n.type = "Other"

        origin = NotifyService()
        assert origin.id is not None
        assert origin.type == origin.DEFAULT_TYPE
        origin.inbox = "http://origin.com/inbox"
        n.origin = origin

        target = NotifyService()
        target.inbox = "http://target.com/inbox"
        n.target = target

        obj = NotifyObject()
        assert obj.id is not None
        assert obj.type is None
        n.object = obj

        actor = NotifyActor()
        assert actor.id is not None
        assert actor.type == actor.DEFAULT_TYPE
        n.actor = actor

        n.in_reply_to = "urn:irt"

        context = NotifyObject()
        assert context.id is not None
        assert context.type is None
        n.context = context

        assert n.id == "urn:whatever"
        assert n.type == "Other"
        assert n.origin.id == origin.id
        assert n.origin.type == origin.DEFAULT_TYPE
        assert n.origin.inbox == "http://origin.com/inbox"
        assert n.target.id == target.id
        assert n.target.type == target.DEFAULT_TYPE
        assert n.target.inbox == "http://target.com/inbox"
        assert n.object.id == obj.id
        assert n.object.type is None
        assert n.actor.id == actor.id
        assert n.actor.type == actor.DEFAULT_TYPE
        assert n.in_reply_to == "urn:irt"
        assert n.context.id == context.id
        assert n.context.type is None

    def test_02_notify_from_fixture(self):
        source = NotifyFixtureFactory.source()
        n = NotifyPattern(source)

        # now check we've got all the source properties
        assert n.id == source["id"]
        assert n.type == source["type"]
        assert isinstance(n.origin, NotifyService)
        assert n.origin.id == source["origin"]["id"]
        assert isinstance(n.object, NotifyObject)
        assert n.object.id == source["object"]["id"]
        assert isinstance(n.target, NotifyService)
        assert n.target.id == source["target"]["id"]
        assert isinstance(n.actor, NotifyActor)
        assert n.actor.id == source["actor"]["id"]
        assert n.in_reply_to == source["inReplyTo"]
        assert isinstance(n.context, NotifyObject)
        assert n.context.id == source["context"]["id"]
        assert isinstance(n.context.item, NotifyItem)
        assert n.context.item.id == source["context"]["ietf:item"]["id"]

        # now check we can rewrite some properties
        n.id = "urn:whatever"
        n.ALLOWED_TYPES = ["Object", "Other"]  # this is a hack to test the setter
        n.type = "Other"
        assert n.id == "urn:whatever"
        assert n.type == "Other"

    def test_03_notify_operations(self):
        n = NotifyPattern()
        with self.assertRaises(ValidationError):
            n.validate()
        assert n.to_jsonld() is not None

        source = NotifyFixtureFactory.source()
        compare = deepcopy(source)
        n = NotifyPattern(source)
        assert n.validate() is True
        assert n.to_jsonld() == compare

    def test_04_accept(self):
        a = Accept()

        source = AcceptFixtureFactory.source()
        compare = deepcopy(source)
        a = Accept(source)
        assert a.validate() is True
        assert a.to_jsonld() == compare

        proptest = self._get_testable_properties(compare)
        self._apply_property_test(proptest, a, AcceptFixtureFactory)

    def test_05_announce_endorsement(self):
        ae = AnnounceEndorsement()
        source = AnnounceEndorsementFixtureFactory.source()
        compare = deepcopy(source)
        ae = AnnounceEndorsement(source)
        assert ae.validate() is True
        assert ae.to_jsonld() == compare

        proptest = self._get_testable_properties(compare)
        self._apply_property_test(proptest, ae, AnnounceEndorsementFixtureFactory)

    def test_07_announce_relationship(self):
        ae = AnnounceRelationship()

        source = AnnounceRelationshipFixtureFactory.source()
        compare = deepcopy(source)
        ae = AnnounceRelationship(source)
        assert ae.validate() is True
        assert ae.to_jsonld() == compare

        proptest = self._get_testable_properties(compare)
        self._apply_property_test(proptest, ae, AnnounceRelationshipFixtureFactory)

    def test_08_announce_review(self):
        ar = AnnounceReview()

        source = AnnounceReviewFixtureFactory.source()
        compare = deepcopy(source)
        ar = AnnounceReview(source)
        assert ar.validate() is True
        assert ar.to_jsonld() == compare

        proptest = self._get_testable_properties(compare)
        self._apply_property_test(proptest, ar, AnnounceReviewFixtureFactory)

    def test_09_announce_service_result(self):
        asr = AnnounceServiceResult()

        source = AnnounceServiceResultFixtureFactory.source()
        compare = deepcopy(source)
        compare["type"] = compare["type"][0]    # because it's a single field, but is a list in the fixture
        asr = AnnounceServiceResult(source)

        assert asr.validate() is True
        assert asr.to_jsonld() == compare

        proptest = self._get_testable_properties(compare)
        self._apply_property_test(proptest, asr, AnnounceServiceResultFixtureFactory)

    def test_10_reject(self):
        rej = Reject()

        source = RejectFixtureFactory.source()
        compare = deepcopy(source)
        rej = Reject(source)
        assert rej.validate() is True
        assert rej.to_jsonld() == compare

        proptest = self._get_testable_properties(compare)
        self._apply_property_test(proptest, rej, RejectFixtureFactory)

    def test_11_request_endorsement(self):
        re = RequestEndorsement()

        source = RequestEndorsementFixtureFactory.source()
        compare = deepcopy(source)
        re = RequestEndorsement(source)

        assert re.validate() is True
        assert re.to_jsonld() == compare

        proptest = self._get_testable_properties(compare)
        self._apply_property_test(proptest, re, RequestEndorsementFixtureFactory)

    def test_13_request_review(self):
        ri = RequestReview()

        source = RequestReviewFixtureFactory.source()
        compare = deepcopy(source)
        ri = RequestReview(source)

        assert ri.validate() is True
        assert ri.to_jsonld() == compare

        proptest = self._get_testable_properties(compare)
        self._apply_property_test(proptest, ri, RequestReviewFixtureFactory)

    def test_14_tentatively_accept(self):
        ta = TentativelyAccept()

        source = TentativelyAcceptFixtureFactory.source()
        compare = deepcopy(source)
        ta = TentativelyAccept(source)

        assert ta.validate() is True
        assert ta.to_jsonld() == compare

        proptest = self._get_testable_properties(compare)
        self._apply_property_test(proptest, ta, TentativelyAcceptFixtureFactory)

    def test_15_tentatively_reject(self):
        ta = TentativelyReject()

        source = TentativelyRejectFixtureFactory.source()
        compare = deepcopy(source)
        ta = TentativelyReject(source)

        assert ta.validate() is True
        assert ta.to_jsonld() == compare

        proptest = self._get_testable_properties(compare)
        self._apply_property_test(proptest, ta, TentativelyRejectFixtureFactory)

    def test_16_unprocessable_notification(self):
        ta = UnprocessableNotification()

        source = UnprocessableNotificationFixtureFactory.source()
        compare = deepcopy(source)
        ta = UnprocessableNotification(source)

        assert ta.validate() is True
        assert ta.to_jsonld() == compare

        proptest = self._get_testable_properties(compare)
        self._apply_property_test(proptest, ta, UnprocessableNotificationFixtureFactory)

    def test_17_undo_offer(self):
        ta = UndoOffer()

        source = UndoOfferFixtureFactory.source()
        compare = deepcopy(source)
        ta = UndoOffer(source)

        assert ta.validate() is True
        assert ta.to_jsonld() == compare

        proptest = self._get_testable_properties(compare)
        self._apply_property_test(proptest, ta, UndoOfferFixtureFactory)

    def test_18_by_ref(self):
        # Create a basic NotifyPatter, and explcitly declare properties by reference to be true (the default)
        n = NotifyPattern(properties_by_reference=True)

        # create an object externally, and confirm that it does not have a specific id
        obj = NotifyObject()
        assert obj.id != "urn:whatever"

        # assign the object to the pattern, then immediately retrieve it by the accessor, and set its
        # id.  We then confirm that both the original object and the pattern's version of the object are
        # *the same* object, and that the id is the same in both references
        n.object = obj
        n.object.id = "urn:whatever"
        assert n.object.id == "urn:whatever"
        assert obj.id == "urn:whatever"

        # Now we confirm that retrieving the record by reference also works.  Create a pattern,
        # and retrieve the object.  Set the ID of the object on the retrieved version, and confirm
        # that the original object also has the same ID (because they are the same, by reference)
        source = RequestReviewFixtureFactory.source()
        n = RequestReview(source, properties_by_reference=True)
        obj = n.object
        obj.id = "urn:whatever"
        assert n.object.id == "urn:whatever"

    def test_19_by_value(self):
        # Create a basic NotifyPatter, and explcitly declare properties by reference to be false.
        # Object should now be copied and passed around by value, so updates to one do not affect
        # the other
        n = NotifyPattern(properties_by_reference=False)

        # create an object externally, and confirm that it does not have a specific id
        obj = NotifyObject()
        assert obj.id != "urn:whatever"

        # assign the object to the pattern, then subsequently update its id.  Retrieve the object
        # from the pattern and confirm that the id change has not propagated
        n.object = obj
        obj.id = "urn:whatever"
        assert n.object.id != "urn:whatever"
        assert obj.id == "urn:whatever"

        # Same test again, but this time pull the object from the pattern first
        source = RequestReviewFixtureFactory.source()
        n = RequestReview(source, properties_by_reference=False)
        obj = n.object
        obj.id = "urn:whatever"
        assert n.object.id != "urn:whatever"




"""
Single file implementation of a test server, showing all the layers of the general
solution in one place.
"""
from flask import Flask, request, make_response
from coarnotify.test.server import settings
from coarnotify.server import COARNotifyServer, COARNotifyServiceBinding, COARNotifyReceipt, COARNotifyServerError
from coarnotify.core.notify import NotifyPattern
import uuid, json, sys, os
from datetime import datetime

def create_app():
    """
    Create the flask app, pulling config from ``settings.py`` then any supplied local config
    in environment variable ``COARNOTIFY_SETTINGS``.

    :return:
    """
    app = Flask(__name__)
    app.config.from_object(settings)
    app.config.from_envvar("COARNOTIFY_SETTINGS", silent=True)
    return app

app = create_app()
"""The global flask app for the test server"""


class COARNotifyServiceTestImpl(COARNotifyServiceBinding):
    """
    Test server implementation of the main service binding
    """
    def notification_received(self, notification: NotifyPattern) -> COARNotifyReceipt:
        """
        Process an incoming notification object in the following way:

        1. Generate a name for the notification based on the timestamp and a random UUID
        2. Write the notification JSON-LD  to a file in the store directory
        3. Return a receipt for the notification using the configured response status and a location pointing to the file

        :param notification:
        :return:
        """
        store = app.config.get("STORE_DIR")
        if not os.path.exists(store):
            print(f"Store directory {store} does not exist, you must create it manually")
            raise COARNotifyServerError(500, "Store directory does not exist")

        now = datetime.utcnow().strftime("%Y%m%d_%H%M%S")
        fn = now + "_" + uuid.uuid4().hex

        with open(f"{store}/{fn}.json", "w") as f:
            f.write(json.dumps(notification.to_jsonld()))

        rstatus = app.config.get("RESPONSE_STATUS", COARNotifyReceipt.CREATED)
        location = f"{request.url_root}inbox/{fn}"

        return COARNotifyReceipt(rstatus, location)


@app.route("/inbox", methods=["POST"])
def inbox():
    """
    Main web entry point.  POST to /inbox to trigger it

    This pulls the notification out of the request as JSON, and sends it to the server
    which will parse it and send it on to the service binding implementation

    When it gets the receipt it will return a blank HTTP response with the appropriate
    status code and Location header

    :return:
    """
    notification = request.json
    server = COARNotifyServer(COARNotifyServiceTestImpl())

    try:
        result = server.receive(notification, validate=app.config.get("VALIDATE_INCOMING", True))
    except COARNotifyServerError as e:
        return make_response(e.message, e.status)

    resp = make_response()
    resp.status_code = result.status
    if result.status == result.CREATED:
        resp.headers["Location"] = result.location
    return resp


def run_server(host=None, port=None, fake_https=False):
    """
    Start the web server using the flask built in server

    :param host:
    :param port:
    :param fake_https:
        if fake_https is True, developer can use https:// to access the server
    :return:
    """
    pycharm_debug = app.config.get('DEBUG_PYCHARM', False)
    if len(sys.argv) > 1:
        if sys.argv[1] == '-d':
            pycharm_debug = True

    if pycharm_debug:
        app.config['DEBUG'] = False
        import pydevd
        pydevd.settrace(app.config.get('DEBUG_PYCHARM_SERVER', 'localhost'),
                        port=app.config.get('DEBUG_PYCHARM_PORT', 6000),
                        stdoutToServer=True, stderrToServer=True)

    # check the store directory exists
    store = app.config.get("STORE_DIR")
    if not os.path.exists(store):
        print(f"Store directory {store} does not exist, you must create it manually")
        exit(1)
    else:
        print(f"Store directory: {store}")

    run_kwargs = {}
    if fake_https:
        run_kwargs['ssl_context'] = 'adhoc'

    host = host or app.config['HOST']
    port = port or app.config['PORT']
    app.run(host=host, debug=app.config['DEBUG'], port=port,
            **run_kwargs)


if __name__ == "__main__":
    run_server()


from unittest import TestCase
from copy import deepcopy

from coarnotify.core.activitystreams2 import ActivityStream, Properties
from coarnotify.test.fixtures.announce_endorsement import AnnounceEndorsementFixtureFactory


class TestActivitystreams(TestCase):
    def test_01_construction(self):
        as2 = ActivityStream()
        assert as2.doc == {}
        assert as2.context == []

        source = AnnounceEndorsementFixtureFactory.source()
        s2 = deepcopy(source)
        s2context = s2.pop("@context")

        as2 = ActivityStream(source)

        assert as2.doc == s2
        assert as2.context == s2context

    def test_02_set_properties(self):
        as2 = ActivityStream()

        # properties that are just basic json
        as2.set_property("random", "value")
        assert as2.doc["random"] == "value"
        assert as2.context == []

        # properties that are in the ASProperties
        as2.set_property(Properties.ID, "value")
        assert as2.doc["id"] == "value"
        assert as2.context == [Properties.ID[1]]

        as2.set_property(Properties.TYPE, "another")
        assert as2.doc["type"] == "another"
        assert as2.context == [Properties.ID[1]]

        # other variations on property namespaces
        as2.set_property(("object", "http://example.com"), "object value")
        as2.set_property(("subject", "http://example.com"), "subject value")
        assert as2.doc["object"] == "object value"
        assert as2.doc["subject"] == "subject value"
        assert as2.context == [Properties.ID[1], "http://example.com"]

        as2.set_property(("foaf:name", ("foaf", "http://xmlns.com/foaf/0.1")), "name value")
        as2.set_property(("foaf:email", ("foaf", "http://xmlns.com/foaf/0.1")), "email value")
        assert as2.doc["foaf:name"] == "name value"
        assert as2.doc["foaf:email"] == "email value"
        assert as2.context == [Properties.ID[1], "http://example.com", {"foaf": "http://xmlns.com/foaf/0.1"}]

    def test_03_get_properties(self):
        as2 = ActivityStream()
        as2.set_property("random", "value")
        as2.set_property(Properties.ID, "id")
        as2.set_property(("object", "http://example.com"), "object value")
        as2.set_property(("foaf:name", ("foaf", "http://xmlns.com/foaf/0.1")), "name value")

        assert as2.get_property("random") == "value"
        assert as2.get_property(Properties.ID) == "id"
        assert as2.get_property(("object", "http://example.com")) == "object value"
        assert as2.get_property("object") == "object value"
        assert as2.get_property(("foaf:name", ("foaf", "http://xmlns.com/foaf/0.1"))) == "name value"
        assert as2.get_property("foaf:name") == "name value"

    def test_04_to_jsonld(self):
        # check we can round trip a document
        source = AnnounceEndorsementFixtureFactory.source()
        s2 = deepcopy(source)
        as2 = ActivityStream(source)
        assert as2.to_jsonld() == s2

        # check we can build a document from scratch and get an expected result
        as2 = ActivityStream()
        as2.set_property("random", "value")
        as2.set_property(Properties.ID, "id")
        as2.set_property(("object", "http://example.com"), "object value")
        as2.set_property(("foaf:name", ("foaf", "http://xmlns.com/foaf/0.1")), "name value")

        expected = {
            "@context": [Properties.ID[1], "http://example.com", {"foaf": "http://xmlns.com/foaf/0.1"}],
            "random": "value",
            "id": "id",
            "object": "object value",
            "foaf:name": "name value"
        }

        assert as2.to_jsonld() == expected



"""
This module contains all the client-specific code for sending notifications
to an inbox and receiving the responses it may return
"""

import json
from typing import Union

from coarnotify.exceptions import NotifyException
from coarnotify.http import RequestsHttpLayer, HttpLayer
from coarnotify.core.notify import NotifyPattern


class NotifyResponse:
    """
    An object representing the response from a COAR Notify inbox.

    This contains the action that was carried out on the server:

    * CREATED - a new resource was created

    * ACCEPTED - the request was accepted, but the resource was not yet created

    In the event that the resource is created, then there will also be a location
    URL which will give you access to the resource
    """
    CREATED = "created"
    ACCEPTED = "accepted"

    def __init__(self, action, location=None):
        """
        Construct a new NotifyResponse object with the action (created or accepted) and the location URL (optional)

        :param action: The action which the server said it took
        :param location: The HTTP URI for the resource that was created (if present)
        """
        self._action = action
        self._location = location

    @property
    def action(self) -> str:
        """The action that was taken, will be one of the constants CREATED or ACCEPTED"""
        return self._action

    @property
    def location(self) -> Union[str, None]:
        """The HTTP URI of the created resource, if present"""
        return self._location


class COARNotifyClient:
    """
    The COAR Notify Client, which is the mechanism through which you will interact with external inboxes.

    If you do not supply an inbox URL at construction you will
    need to supply it via the ``inbox_url`` setter, or when you send a notification

    :param inbox_url:   HTTP URI of the inbox to communicate with by default
    :param http_layer:  An implementation of the HttpLayer interface to use for sending HTTP requests.
                        If not provided, the default implementation will be used based on ``requests``
    """
    def __init__(self, inbox_url: str = None, http_layer: HttpLayer = None):
        self._inbox_url = inbox_url
        self._http = http_layer if http_layer is not None else RequestsHttpLayer()

    @property
    def inbox_url(self) -> Union[str, None]:
        """The HTTP URI of the inbox to communicate with by default"""
        return self._inbox_url

    @inbox_url.setter
    def inbox_url(self, value: str):
        """Set the HTTP URI of the inbox to communicate with by default"""
        self._inbox_url = value

    def send(self, notification: NotifyPattern, inbox_url: str = None, validate: bool = True) -> NotifyResponse:
        """
        Send the given notification to the inbox.  If no inbox URL is provided, the default inbox URL will be used.

        :param notification: The notification object (from the models provided, or a subclass you have made of the NotifyPattern class)
        :param inbox_url: The HTTP URI to send the notification to.  Omit if using the default inbox_url supplied in the constructor.
                            If it is omitted, and no value is passed here then we will also look in the ``target.inbox`` property of the notification
        :param validate: Whether to validate the notification before sending.  If you are sure the notification is valid, you can set this to False
        :return: a NotifyResponse object representing the response from the server
        """
        if inbox_url is None:
            inbox_url = self._inbox_url
        if inbox_url is None:
            inbox_url = notification.target.inbox
        if inbox_url is None:
            raise ValueError("No inbox URL provided at the client, method, or notification level")

        if validate:
            if not notification.validate():
                raise NotifyException("Attempting to send invalid notification; to override set validate=False when calling this method")

        resp = self._http.post(inbox_url,
                        data=json.dumps(notification.to_jsonld()),
                        headers={"Content-Type": "application/ld+json;profile=\"https://www.w3.org/ns/activitystreams\""}
                        )

        if resp.status_code == 201:
            return NotifyResponse(NotifyResponse.CREATED, location=resp.header("Location"))
        elif resp.status_code == 202:
            return NotifyResponse(NotifyResponse.ACCEPTED)

        raise NotifyException("Unexpected response: %s" % resp.status_code)



from copy import deepcopy


class URIFixtureFactory:
    @classmethod
    def generate(cls, schemes=None, hosts=None, ports=None, paths=None, queries=None, fragments=None):
        schemes = schemes if schemes is not None else deepcopy(DEFAULT_SCHEMES)
        hosts = hosts if hosts is not None else deepcopy(DEFAULT_HOSTS)
        ports = ports if ports is not None else deepcopy(DEFAULT_PORTS)
        paths = paths if paths is not None else deepcopy(DEFAULT_PATHS)
        queries = queries if queries is not None else deepcopy(DEFAULT_QUERIES)
        fragments = fragments if fragments is not None else deepcopy(DEFAULT_FRAGMENTS)

        uris = []
        for scheme in schemes:
            for host in hosts:
                for port in ports:
                    for path in paths:
                        for query in queries:
                            for fragment in fragments:
                                uris.append(cls.generate_uri(scheme, host, port, path, query, fragment))

        return uris

    @classmethod
    def generate_uri(cls, scheme, host, port, path, query, fragment):
        # account for port numbers and IPv6 addresses
        if host is not None and ":" in host and port is not None and port != "":
            host = f"[{host}]"
        url = scheme + "://" if scheme is not None and scheme != "" else ""
        url += host if host is not None else ""
        url += ":" + port if port is not None and port != "" else ""
        url += path if path is not None else ""
        url += "?" + query if query is not None and query != "" else ""
        url += "#" + fragment if fragment is not None and fragment != "" else ""
        return url


DEFAULT_SCHEMES = [
    "http",
    "https"
]

DEFAULT_HOSTS = [
    "example.com",
    "localhost",
    "192.168.0.1",
    "2001:db8::7"
]

DEFAULT_PORTS = [
    "",
    "80",
    "8080"
]

DEFAULT_PATHS = [
    "",
    "/",
    "/path",
    "/path/to/file"
]

DEFAULT_QUERIES = [
    "",
    "query",
    "query=string&o=1",
]

DEFAULT_FRAGMENTS = [
    "",
    "fragment"
]


from unittest import TestCase

from coarnotify.client import COARNotifyClient
from coarnotify.patterns import AnnounceEndorsement
from coarnotify.test.fixtures import AnnounceEndorsementFixtureFactory
from coarnotify.test.mocks.http import MockHttpResponse, MockHttpLayer


class TestClient(TestCase):
    def test_01_construction(self):
        client = COARNotifyClient()
        assert client.inbox_url is None

        client = COARNotifyClient("http://example.com/inbox")
        assert client.inbox_url == "http://example.com/inbox"

        client = COARNotifyClient(http_layer=MockHttpLayer())
        client = COARNotifyClient("http://example.com/inbox", MockHttpLayer())

    def test_02_created_response(self):
        client = COARNotifyClient("http://example.com/inbox", MockHttpLayer(
            status_code=201,
            location="http://example.com/location"
        ))
        source = AnnounceEndorsementFixtureFactory.source()
        ae = AnnounceEndorsement(source)
        resp = client.send(ae)
        assert resp.action == resp.CREATED
        assert resp.location == "http://example.com/location"


    def test_03_accepted_response(self):
        client = COARNotifyClient("http://example.com/inbox", MockHttpLayer(
            status_code=202
        ))
        source = AnnounceEndorsementFixtureFactory.source()
        ae = AnnounceEndorsement(source)
        resp = client.send(ae)
        assert resp.action == resp.ACCEPTED
        assert resp.location is None



"""
This module contains all the test infrastructure.

For the purposes of conciseness, the test code is mostly excluded from the auto-documentation.

The test code is structured as follows:

- `fixtures` contains test fixtures and factories for generating fixtures
- `integration` contains integration tests for the library.  These depend on a running server, and there is a test server provided
- `mocks` contains mock objects for testing
- `server` contains the test server, and documentation for this is generated and available on the docsite
- `unit` contains unit tests for the library

For information on running the integration and unit tests, see :doc:`/dev`.
"""


from coarnotify.test.fixtures.base_fixture_factory import BaseFixtureFactory
from coarnotify.test.fixtures.accept import AcceptFixtureFactory
from coarnotify.test.fixtures.announce_endorsement import AnnounceEndorsementFixtureFactory
from coarnotify.test.fixtures.announce_relationship import AnnounceRelationshipFixtureFactory
from coarnotify.test.fixtures.announce_review import AnnounceReviewFixtureFactory
from coarnotify.test.fixtures.announce_service_result import AnnounceServiceResultFixtureFactory
from coarnotify.test.fixtures.reject import RejectFixtureFactory
from coarnotify.test.fixtures.request_endorsement import RequestEndorsementFixtureFactory
from coarnotify.test.fixtures.uris import URIFixtureFactory
from coarnotify.test.fixtures.request_review import RequestReviewFixtureFactory
from coarnotify.test.fixtures.tentatively_accept import TentativelyAcceptFixtureFactory
from coarnotify.test.fixtures.tentatively_reject import TentativelyRejectFixtureFactory
from coarnotify.test.fixtures.unprocessable_notification import UnprocessableNotificationFixtureFactory
from coarnotify.test.fixtures.undo_offer import UndoOfferFixtureFactory


"""
Pattern to represent the Announce Review notification
https://coar-notify.net/specification/1.0.0/announce-review/
"""
from coarnotify.core.notify import NotifyPattern, NotifyTypes, NotifyObject, NotifyItem, NotifyProperties
from coarnotify.core.activitystreams2 import ActivityStreamsTypes, Properties
from coarnotify.exceptions import ValidationError

from typing import Union

__all__ = ["AnnounceReview", "AnnounceReviewContext", "AnnounceReviewItem", "AnnounceReviewObject"]

class AnnounceReview(NotifyPattern):
    """
    Class to represent Announce Review pattern
    """
    TYPE = [ActivityStreamsTypes.ANNOUNCE, NotifyTypes.REVIEW_ACTION]
    """ Announce Review type, including Acitivity Streams Announce and Notify Review Action """

    @property
    def object(self) -> Union["AnnounceReviewObject", None]:
        """
        Custom getter to retrieve Announce Review object

        :return: Announce Review Object
        """
        o = self.get_property(Properties.OBJECT)
        if o is not None:
            return AnnounceReviewObject(o,
                                validate_stream_on_construct=False,
                                validate_properties=self.validate_properties,
                                validators=self.validators,
                                validation_context=Properties.OBJECT,
                                properties_by_reference=self._properties_by_reference)
        return None

    @property
    def context(self) -> Union["AnnounceReviewContext", None]:
        """
        Custom getter to retrieve AnnounceReview Context

        :return: AnnounceReviewContext
        """
        c = self.get_property(Properties.CONTEXT)
        if c is not None:
            return AnnounceReviewContext(c,
                                      validate_stream_on_construct=False,
                                      validate_properties=self.validate_properties,
                                      validators=self.validators,
                                      validation_context=Properties.CONTEXT,
                                      properties_by_reference=self._properties_by_reference)
        return None

    def validate(self) -> bool:
        """
        Extends the base validation to make `context` required

        :return: ``True`` if valid, otherwise raises :py:class:`coarnotify.exceptions.ValidationError`
        """
        ve = ValidationError()
        try:
            super(NotifyPattern, self).validate()
        except ValidationError as superve:
            ve = superve

        self.required_and_validate(ve, Properties.CONTEXT, self.context)

        if ve.has_errors():
            raise ve

        return True

class AnnounceReviewContext(NotifyObject):
    """
    Custom Context for Announce Review, specifically to return custom
    Announce Review Item
    """
    @property
    def item(self) -> Union["AnnounceReviewItem", None]:
        """
        Custom getter to retrieve AnnounceReviewItem

        :return: AnnounceReviewItem
        """
        i = self.get_property(NotifyProperties.ITEM)
        if i is not None:
            return AnnounceReviewItem(i,
                              validate_stream_on_construct=False,
                              validate_properties=self.validate_properties,
                              validators=self.validators,
                              validation_context=NotifyProperties.ITEM,
                              properties_by_reference=self._properties_by_reference)
        return None


class AnnounceReviewItem(NotifyItem):
    """
    Custom AnnounceReviewItem which provides additional validation over the basic NotifyItem
    """

    def validate(self) -> bool:
        """
        In addition to the base validator, this:

        * Reintroduces type validation
        * make ``mediaType`` a required field

        :return: ``True`` if valid, else raises a ValidationError
        """
        ve = ValidationError()
        try:
            super(AnnounceReviewItem, self).validate()
        except ValidationError as superve:
            ve = superve

        self.required_and_validate(ve, Properties.TYPE, self.type)
        self.required(ve, NotifyProperties.MEDIA_TYPE, self.media_type)

        if ve.has_errors():
            raise ve
        return True


class AnnounceReviewObject(NotifyObject):
    """
    Custom Announce Review Object to apply custom validation for this pattern
    """

    def validate(self) -> bool:
        """
        In addition to the base validator this:

        * Makes type required

        :return: ``True`` if valid, else raises ValidationError
        """
        ve = ValidationError()
        try:
            super(AnnounceReviewObject, self).validate()
        except ValidationError as superve:
            ve = superve

        self.required_and_validate(ve, Properties.TYPE, self.type)

        if ve.has_errors():
            raise ve

        return True


from copy import deepcopy

from coarnotify.test.fixtures import BaseFixtureFactory


class AnnounceReviewFixtureFactory(BaseFixtureFactory):
    @classmethod
    def source(cls, copy=True):
        if copy:
            return deepcopy(ANNOUNCE_REVIEW)
        return ANNOUNCE_REVIEW


ANNOUNCE_REVIEW = {
    "@context": [
        "https://www.w3.org/ns/activitystreams",
        "https://coar-notify.net"
    ],
    "actor": {
        "id": "https://review-service.com",
        "name": "Review Service",
        "type": "Service"
    },
    "context": {
        "id": "https://research-organisation.org/repository/preprint/201203/421/"
    },
    "id": "urn:uuid:94ecae35-dcfd-4182-8550-22c7164fe23f",
    "inReplyTo": "urn:uuid:0370c0fb-bb78-4a9b-87f5-bed307a509dd",
    "object": {
        "id": "https://review-service.com/review/geo/202103/0021",
        "ietf:cite-as": "https://doi.org/10.3214/987654",
        "type": [
            "Page",
            "sorg:Review"
        ]
    },
    "origin": {
        "id": "https://review-service.com/system",
        "inbox": "https://review-service.com/inbox/",
        "type": "Service"
    },
    "target": {
        "id": "https://generic-service.com/system",
        "inbox": "https://generic-service.com/system/inbox/",
        "type": "Service"
    },
    "type": [
        "Announce",
        "coar-notify:ReviewAction"
    ]
}




from unittest import TestCase

from coarnotify.core.notify import NotifyPattern, NotifyService, NotifyObject
from coarnotify.patterns import (
    Accept,
    AnnounceEndorsement,
    AnnounceRelationship,
    AnnounceReview,
    AnnounceServiceResult,
    RequestEndorsement,
    RequestReview,
    TentativelyAccept,
    TentativelyReject,
    UnprocessableNotification,
    UndoOffer,
    Reject
)
from coarnotify.test.fixtures.notify import NotifyFixtureFactory
from coarnotify.test.fixtures import (
    AcceptFixtureFactory,
    AnnounceEndorsementFixtureFactory,
    AnnounceRelationshipFixtureFactory,
    AnnounceReviewFixtureFactory,
    AnnounceServiceResultFixtureFactory,
    RequestReviewFixtureFactory,
    RequestEndorsementFixtureFactory,
    URIFixtureFactory,
    TentativelyAcceptFixtureFactory,
    TentativelyRejectFixtureFactory,
    UnprocessableNotificationFixtureFactory,
    UndoOfferFixtureFactory,
    RejectFixtureFactory
)

from coarnotify.exceptions import ValidationError
from coarnotify.core.activitystreams2 import Properties
from coarnotify.core.notify import NotifyProperties
from coarnotify import validate
from coarnotify.validate import Validator


class TestValidate(TestCase):
    def test_01_structural_empty(self):
        n = NotifyPattern()
        n.id = None     # these are automatically set, so remove them to trigger validation
        n.type = None
        with self.assertRaises(ValidationError) as ve:
            n.validate()

        errors = ve.exception.errors
        assert Properties.ID in errors
        assert Properties.TYPE in errors
        assert Properties.OBJECT in errors
        assert Properties.TARGET in errors
        assert Properties.ORIGIN in errors

    def test_02_structural_basic(self):
        n = NotifyPattern()
        with self.assertRaises(ValidationError) as ve:
            n.validate()

        errors = ve.exception.errors
        assert Properties.ID not in errors
        assert Properties.TYPE not in errors
        assert Properties.OBJECT in errors
        assert Properties.TARGET in errors
        assert Properties.ORIGIN in errors

    def test_03_structural_valid_document(self):
        n = NotifyPattern()
        n.target = NotifyFixtureFactory.target()
        n.origin = NotifyFixtureFactory.origin()
        n.object = NotifyFixtureFactory.object()

        assert n.validate() is True

    def test_04_structural_invalid_nested(self):
        n = NotifyPattern()
        n.target = NotifyService({"whatever": "value"}, validate_stream_on_construct=False)
        n.origin = NotifyService({"another": "junk"}, validate_stream_on_construct=False)
        n.object = NotifyObject({"yet": "more"}, validate_stream_on_construct=False)

        with self.assertRaises(ValidationError) as ve:
            n.validate()

        errors = ve.exception.errors
        assert Properties.ID not in errors
        assert Properties.TYPE not in errors
        assert Properties.OBJECT not in errors  # the object is present, and will acquire an id, so will not be in the errors
        assert Properties.TARGET in errors
        assert Properties.ORIGIN in errors

        # These are no longer causing validation errors, as the index field is not being checked

        # target = errors[Properties.TARGET]
        # assert len(target.get("errors")) == 0
        # assert target.get("nested") is not None
        # assert NotifyProperties.INBOX in target.get("nested")

        # origin = errors[Properties.ORIGIN]
        # assert len(origin.get("errors")) == 0
        # assert origin.get("nested") is not None
        # assert NotifyProperties.INBOX in origin.get("nested")

    def test_05_validation_modes(self):
        valid = NotifyFixtureFactory.source()
        n = NotifyPattern(stream=valid, validate_stream_on_construct=True)

        invalid = NotifyFixtureFactory.source()
        invalid["id"] = "http://example.com/^path"
        with self.assertRaises(ValidationError) as ve:
            n = NotifyPattern(stream=invalid, validate_stream_on_construct=True)
        assert ve.exception.errors.get(Properties.ID) is not None

        valid = NotifyFixtureFactory.source()
        n = NotifyPattern(stream=valid, validate_stream_on_construct=False)

        invalid = NotifyFixtureFactory.source()
        invalid["id"] = "http://example.com/^path"
        n = NotifyPattern(stream=invalid, validate_stream_on_construct=False)

        n = NotifyPattern(validate_properties=False)
        n.id = "urn:uuid:4fb3af44-d4f8-4226-9475-2d09c2d8d9e0"  # valid
        n.id = "http://example.com/^path"   # invalid

        with self.assertRaises(ValidationError) as ve:
            n.validate()
        assert ve.exception.errors.get(Properties.ID) is not None

    def test_06_validate_id_property(self):
        n = NotifyPattern()
        # test the various ways it can fail:
        with self.assertRaises(ValueError) as ve:
            n.id = "9whatever:none"
        assert ve.exception.args[0] == "Invalid URI scheme `9whatever`"

        with self.assertRaises(ValueError) as ve:
            n.id = "http://wibble/stuff"
        assert ve.exception.args[0] == "Invalid URI authority `wibble`"

        with self.assertRaises(ValueError) as ve:
            n.id = "http://example.com/^path"
        assert ve.exception.args[0] == "Invalid URI path `/^path`"

        with self.assertRaises(ValueError) as ve:
            n.id = "http://example.com/path/here/?^=what"
        assert ve.exception.args[0] == "Invalid URI query `^=what`"

        with self.assertRaises(ValueError) as ve:
            n.id = "http://example.com/path/here/?you=what#^frag"
        assert ve.exception.args[0] == "Invalid URI fragment `^frag`"

        # test a bunch of successful ones

        # These ones taken from wikipedia
        n.id = "https://john.doe@www.example.com:1234/forum/questions/?tag=networking&order=newest#top"
        n.id = "https://john.doe@www.example.com:1234/forum/questions/?tag=networking&order=newest#:~:text=whatever"
        n.id = "ldap://[2001:db8::7]/c=GB?objectClass?one"
        n.id = "mailto:John.Doe@example.com"
        n.id = "news:comp.infosystems.www.servers.unix"
        n.id = "tel:+1-816-555-1212"
        n.id = "telnet://192.0.2.16:80/"
        n.id = "urn:oasis:names:specification:docbook:dtd:xml:4.1.2"

        # these ones taken from the spec
        n.id = "urn:uuid:4fb3af44-d4f8-4226-9475-2d09c2d8d9e0"
        n.id = "https://generic-service.com/system"
        n.id = "https://generic-service.com/system/inbox/"

    def test_07_validate_url(self):
        urls = URIFixtureFactory.generate(schemes=["http", "https"])
        # print(urls)

        for url in urls:
            # print(url)
            assert validate.url(None, url) is True

        with self.assertRaises(ValueError):
            validate.url(None, "ftp://example.com")
        with self.assertRaises(ValueError):
            validate.url(None, "http:/example.com")
        with self.assertRaises(ValueError):
            validate.url(None, "http://domain/path")
        with self.assertRaises(ValueError):
            validate.url(None, "http://example.com/path^wrong")

    def test_08_one_of(self):
        values = ["a", "b", "c"]
        validator = validate.one_of(values)
        assert validator(None, "a") is True
        assert validator(None, "b") is True
        assert validator(None, "c") is True

        with self.assertRaises(ValueError):
            validator(None, "d")

        with self.assertRaises(ValueError):
            # one_of expects a singular value, it does not do lists
            validator(None, ["a", "b"])

    def test_09_contains(self):
        validator = validate.contains("a")
        assert validator(None, ["a", "b", "c"]) is True

        with self.assertRaises(ValueError):
            validator(None, ["b", "c", "d"])

    def test_10_at_least_one_of(self):
        values = ["a", "b", "c"]
        validator = validate.at_least_one_of(values)
        assert validator(None, "a") is True
        assert validator(None, "b") is True
        assert validator(None, "c") is True

        with self.assertRaises(ValueError):
            validator(None, "d")

        # at_least_one_of can take a list and validate each one against the global criteria
        assert validator(None, ["a", "d"]) is True

    ########################################
    ## validation methods for specific patterns

    def _base_validate(self, a):
        # now try to apply invalid values to it
        with self.assertRaises(ValueError):
            a.id = "not a uri"

        with self.assertRaises(ValueError):
            a.in_reply_to = "not a uri"

        with self.assertRaises(ValueError):
            # not an HTTP URI
            a.origin.id = "urn:uuid:4fb3af44-d4f8-4226-9475-2d09c2d8d9e0"

        with self.assertRaises(ValueError):
            a.origin.inbox = "not a uri"

        with self.assertRaises(ValueError):
            # not an HTTP URI
            a.target.id = "urn:uuid:4fb3af44-d4f8-4226-9475-2d09c2d8d9e0"

        with self.assertRaises(ValueError):
            a.target.inbox = "not a uri"

        with self.assertRaises(ValueError):
            a.type = "NotAValidType"

    def _actor_validate(self, a):
        with self.assertRaises(ValueError):
            a.actor.id = "not a uri"

        with self.assertRaises(ValueError):
            a.actor.type = "NotAValidType"

    def _object_validate(self, a):
        with self.assertRaises(ValueError):
            a.object.id = "not a uri"

        with self.assertRaises(ValueError):
            a.object.cite_as = "urn:uuid:4fb3af44-d4f8-4226-9475-2d09c2d8d9e0"

    def _context_validate(self, a):
        with self.assertRaises(ValueError):
            a.context.id = "not a uri"

        with self.assertRaises(ValueError):
            a.context.type = "NotAValidType"

        with self.assertRaises(ValueError):
            a.context.cite_as = "urn:uuid:4fb3af44-d4f8-4226-9475-2d09c2d8d9e0"

    def test_11_accept_validate(self):
        # make a valid one
        source = AcceptFixtureFactory.source()
        a = Accept(source)

        self._base_validate(a)

        # now make one with fully invalid data
        isource = AcceptFixtureFactory.invalid()
        with self.assertRaises(ValidationError) as ve:
            a = Accept(isource)

    def test_12_announce_endorsement_validate(self):
        # make a valid one
        source = AnnounceEndorsementFixtureFactory.source()
        a = AnnounceEndorsement(source)

        self._base_validate(a)

        with self.assertRaises(ValueError):
            # one of the required types, but not both of them
            a.type = "Announce"

        self._actor_validate(a)
        self._object_validate(a)
        self._context_validate(a)

        # now make one with fully invalid data
        isource = AnnounceEndorsementFixtureFactory.invalid()
        with self.assertRaises(ValidationError) as ve:
            a = AnnounceEndorsement(isource)

    def test_13_tentative_accept_validate(self):
        # make a valid one
        source = TentativelyAcceptFixtureFactory.source()
        a = TentativelyAccept(source)

        self._base_validate(a)
        self._actor_validate(a)
        # omit object validation, as the nested object here is a full notification in its own right
        # self._object_validate(a)

        # now make one with fully invalid data
        isource = TentativelyAcceptFixtureFactory.invalid()
        with self.assertRaises(ValidationError) as ve:
            a = TentativelyAccept(isource)

    def test_14_tentative_reject_validate(self):
        # make a valid one
        source = TentativelyRejectFixtureFactory.source()
        a = TentativelyReject(source)

        self._base_validate(a)
        self._actor_validate(a)
        # omit object validation, as the nested object here is a full notification in its own right
        # self._object_validate(a)

        # now make one with fully invalid data
        isource = TentativelyRejectFixtureFactory.invalid()
        with self.assertRaises(ValidationError) as ve:
            a = TentativelyReject(isource)

    def test_15_unprocessable_notification_validate(self):
        # make a valid one
        source = UnprocessableNotificationFixtureFactory.source()
        a = UnprocessableNotification(source)

        self._base_validate(a)
        self._actor_validate(a)
        # self._object_validate(a)

        # now make one with fully invalid data
        isource = UnprocessableNotificationFixtureFactory.invalid()
        with self.assertRaises(ValidationError) as ve:
            a = UnprocessableNotification(isource)

    def test_16_undo_offer_validate(self):
        # make a valid one
        source = UndoOfferFixtureFactory.source()
        a = UndoOffer(source)

        self._base_validate(a)
        self._actor_validate(a)
        # self._object_validate(a)

        # now make one with fully invalid data
        isource = UndoOfferFixtureFactory.invalid()
        with self.assertRaises(ValidationError) as ve:
            a = UnprocessableNotification(isource)


    def test_17_announce_review_validate(self):
        # make a valid one
        source = AnnounceReviewFixtureFactory.source()
        a = AnnounceReview(source)

        self._base_validate(a)

        with self.assertRaises(ValueError):
            # one of the required types, but not both of them
            a.type = "Offer"

        self._actor_validate(a)
        self._object_validate(a)

        # now make one with fully invalid data
        isource = AnnounceReviewFixtureFactory.invalid()
        with self.assertRaises(ValidationError) as ve:
            a = AnnounceReview(isource)

    def test_18_request_endorsement_validate(self):
        # make a valid one
        source = RequestEndorsementFixtureFactory.source()
        a = RequestEndorsement(source)

        self._base_validate(a)
        self._actor_validate(a)
        self._object_validate(a)

        with self.assertRaises(ValueError):
            # one of the required types, but not both of them
            a.type = "Offer"

        # now make one with fully invalid data
        isource = RequestEndorsementFixtureFactory.invalid()
        with self.assertRaises(ValidationError) as ve:
            a = RequestEndorsement(isource)

    def test_19_request_review_validate(self):
        # make a valid one
        source = RequestReviewFixtureFactory.source()
        a = RequestReview(source)

        self._base_validate(a)
        self._actor_validate(a)
        self._object_validate(a)

        with self.assertRaises(ValueError):
            # one of the required types, but not both of them
            a.type = "Offer"

        # now make one with fully invalid data
        isource = RequestReviewFixtureFactory.invalid()
        with self.assertRaises(ValidationError) as ve:
            a = RequestReview(isource)

    def test_20_reject_validate(self):
        # make a valid one
        source = RejectFixtureFactory.source()
        a = Reject(source)

        self._base_validate(a)
        self._actor_validate(a)
        # self._object_validate(a)

        # now make one with fully invalid data
        isource = RejectFixtureFactory.invalid()
        with self.assertRaises(ValidationError) as ve:
            a = Reject(isource)

    def test_21_announce_relationship_validate(self):
        # make a valid one
        source = AnnounceRelationshipFixtureFactory.source()
        a = AnnounceRelationship(source)

        self._base_validate(a)
        self._actor_validate(a)
        self._object_validate(a)
        self._context_validate(a)

        with self.assertRaises(ValueError):
            # one of the required types, but not both of them
            a.type = "Announce"

        # now make one with fully invalid data
        isource = AnnounceRelationshipFixtureFactory.invalid()
        with self.assertRaises(ValidationError) as ve:
            a = AnnounceRelationship(isource)

    def test_21_announce_service_result_validate(self):
        # make a valid one
        source = AnnounceServiceResultFixtureFactory.source()
        a = AnnounceServiceResult(source)

        self._base_validate(a)
        self._actor_validate(a)
        self._object_validate(a)
        self._context_validate(a)

        # now make one with fully invalid data
        isource = AnnounceServiceResultFixtureFactory.invalid()
        with self.assertRaises(ValidationError) as ve:
            a = AnnounceServiceResult(isource)

    def test_22_add_rules(self):
        rules = {
            Properties.ID: {
                "default": validate.absolute_uri,
                "context": {
                    Properties.CONTEXT: {
                        "default": validate.url
                    },
                    Properties.ORIGIN: {
                        "default": validate.url
                    },
                    Properties.TARGET: {
                        "default": validate.url
                    },
                    NotifyProperties.ITEM: {
                        "default": validate.url
                    }
                }
            },
            Properties.TYPE: {
                "default": validate.type_checker,
            }
        }

        v = Validator(rules)

        update = {
            Properties.ID: {
                "default": validate.url
            },
            Properties.TYPE: {
                "context": {
                    Properties.CONTEXT: {
                        "default": validate.url
                    }
                }
            },
            Properties.ACTOR : {
                "default": validate.url
            }
        }

        v.add_rules(update)

        rules = v.rules()
        assert rules[Properties.ID]["default"] == validate.url
        assert rules[Properties.ID]["context"][Properties.CONTEXT]["default"] == validate.url
        assert rules[Properties.TYPE]["default"] == validate.type_checker
        assert rules[Properties.TYPE]["context"][Properties.CONTEXT]["default"] == validate.url
        assert rules[Properties.ACTOR]["default"] == validate.url



from copy import deepcopy

from coarnotify.test.fixtures import BaseFixtureFactory


class UnprocessableNotificationFixtureFactory(BaseFixtureFactory):
    @classmethod
    def source(cls, copy=True):
        if copy:
            return deepcopy(UNPROCESSABLE_NOTIFICATION)
        return UNPROCESSABLE_NOTIFICATION

    @classmethod
    def invalid(cls):
        source = cls.source()
        cls._base_invalid(source)
        del source["summary"]
        return source


UNPROCESSABLE_NOTIFICATION = {
    "@context": [
        "https://www.w3.org/ns/activitystreams",
        "https://coar-notify.net"
    ],
    "actor": {
        "id": "https://generic-service-1.com",
        "name": "Generic Service",
        "type": "Service"
    },
    "id": "urn:uuid:49dae4d9-4a16-4dcf-8ae0-a0cef139254c",
    "inReplyTo": "urn:uuid:0370c0fb-bb78-4a9b-87f5-bed307a509dd",
    "object": {
        "id": "urn:uuid:0370c0fb-bb78-4a9b-87f5-bed307a509dd"
    },
    "origin": {
        "id": "https://some-organisation.org",
        "inbox": "https://some-organisation.org/inbox/",
        "type": "Service"
    },
    "summary": "Unable to process URL: http://www.example.com/broken-url - returns HTTP error 404",
    "target": {
        "id": "https://generic-service.com/system",
        "inbox": "https://generic-service.com/system/inbox/",
        "type": "Service"
    },
    "type": [
        "Flag",
        "coar-notify:UnprocessableNotification"
    ]
}


"""
This is the base of the `coarnotifypy` module.

In here you will find
a full set of model objects for all the Notify Patterns documented in
https://coar-notify.net/specification/1.0.1/

You will also find a client library that will allow you to send notifications
to an inbox, and a server library that will allow you to write a service
binding to your own systems to receive notifications via an inbox.

There are also unit tests demonstrating the various features of the system,
integration tests which can be run against a remote inbox, and a
stand-alone inbox you can use for local testing.
"""

__version__ = "1.0.1.3"


"""
Pattern to represent the Unprocessable Notification notification
https://coar-notify.net/specification/1.0.0/unprocessable/
"""

from coarnotify.core.notify import NotifyPattern, SummaryMixin, NotifyTypes
from coarnotify.core.activitystreams2 import ActivityStreamsTypes, Properties
from coarnotify.exceptions import ValidationError

__all__ = ["UnprocessableNotification"]

class UnprocessableNotification(NotifyPattern, SummaryMixin):
    """
    Class to represent the Unprocessable Notification notification
    """
    TYPE = [ActivityStreamsTypes.FLAG, NotifyTypes.UNPROCESSABLE_NOTIFICATION]
    """Unprocessable Notification types, including an ActivityStreams Flag and a COAR Notify Unprocessable Notification"""

    def validate(self) -> bool:
        """
        In addition to the base validation apply the following constraints:

        * The ``inReplyTo`` property is required
        * The ``summary`` property is required

        :return:
        """
        ve = ValidationError()
        try:
            super(UnprocessableNotification, self).validate()
        except ValidationError as superve:
            ve = superve

        # Technically, no need to validate the value, as this is handled by the superclass,
        # but leaving it in for completeness
        self.required_and_validate(ve, Properties.IN_REPLY_TO, self.in_reply_to)
        self.required(ve, Properties.SUMMARY, self.summary)

        if ve.has_errors():
            raise ve

        return True



from copy import deepcopy
from coarnotify.test.fixtures import BaseFixtureFactory

class AnnounceRelationshipFixtureFactory(BaseFixtureFactory):
    @classmethod
    def source(cls, copy=True):
        if copy:
            return deepcopy(ANNOUNCE_RELATIONSHIP)
        return ANNOUNCE_RELATIONSHIP

    @classmethod
    def invalid(cls):
        source = cls.source()
        cls._base_invalid(source)
        cls._actor_invalid(source)
        cls._object_invalid(source)
        cls._context_invalid(source)
        return source


ANNOUNCE_RELATIONSHIP = {
    "@context": [
        "https://www.w3.org/ns/activitystreams",
        "https://coar-notify.net"
    ],
    "actor": {
        "id": "https://research-organisation.org",
        "name": "Research Organisation",
        "type": "Organization"
    },
    "context": {
        "id": "https://another-research-organisation.org/repository/datasets/item/201203421/",
        "ietf:cite-as": "https://doi.org/10.5555/999555666",
        "ietf:item": {
            "id": "https://another-research-organisation.org/repository/datasets/item/201203421/data_archive.zip",
            "mediaType": "application/zip",
            "type": [
                "Object",
                "sorg:Dataset"
            ]
        },
        "type": [
            "Page",
            "sorg:AboutPage"
        ]
    },
    "id": "urn:uuid:94ecae35-dcfd-4182-8550-22c7164fe23f",
    "object": {
        "as:object": "https://another-research-organisation.org/repository/datasets/item/201203421/",
        "as:relationship": "http://purl.org/vocab/frbr/core#supplement",
        "as:subject": "https://research-organisation.org/repository/item/201203/421/",
        "id": "urn:uuid:74FFB356-0632-44D9-B176-888DA85758DC",
        "type": "Relationship"
    },
    "origin": {
        "id": "https://research-organisation.org/repository",
        "inbox": "https://research-organisation.org/inbox/",
        "type": "Service"
    },
    "target": {
        "id": "https://another-research-organisation.org/repository",
        "inbox": "https://another-research-organisation.org/inbox/",
        "type": "Service"
    },
    "type": [
        "Announce",
        "coar-notify:RelationshipAction"
    ]
}



"""
Module for custom exceptions
"""

class NotifyException(Exception):
    """
    Base class for all exceptions in the coarnotifypy library
    """
    pass


class ValidationError(NotifyException):
    """
    Exception class for validation errors.

    :param errors: a dictionary of errors to construct the exception around.  See below for the details
        of its structure

    This class is designed to be thrown and caught and to collect validation errors
    as it passed through the validation pipeline.

    For example an object validator may do something like this:

    .. code-block:: python

        def validate():
            ve = ValidationError()
            ve.add_error(prop_name, validate.REQUIRED_MESSAGE.format(x=pn))
            if ve.has_errors():
                raise ve
            return True

    If this is called by a subclass which is also validating, then this may be used
    like this:

    .. code-block:: python

        def validate():
            ve = ValidationError()
            try:
                super(ClassName, self).validate()
            except ValidationError as superve:
                ve = superve

            ve.add_error(prop_name, validate.REQUIRED_MESSAGE.format(x=pn))
            if ve.has_errors():
                raise ve
            return True

    By the time the ValidationError is finally raised to the top, it will contain
    all the validation errors from the various levels of validation that have been
    performed.

    The errors are stored as a multi-level dictionary with the keys at the top level
    being the fields in the data structure which have errors, and within the value
    for each key there are two possible keys:

    * errors: a list of error messages for this field
    * nested: a dictionary of further errors for nested fields

    .. code-block:: python

        {
            "key1": {
                "errors": ["error1", "error2"],
                "nested: {
                    "key2": {
                        errors: ["error3"]
                    }
                }
            }
        }

    """
    def __init__(self, errors: dict=None):
        super().__init__()
        self._errors = errors if errors is not None else {}

    @property
    def errors(self) -> dict:
        """The dictionary of errors"""
        return self._errors

    def add_error(self, key: str, value: str):
        """
        Record an error on the supplied ``key`` with the message ``value``

        :param key: the key for which an error is to be recorded
        :param value:   the error message
        :return:
        """
        if key not in self._errors:
            self._errors[key] = {"errors": []}
        self._errors[key]["errors"].append(value)

    def add_nested_errors(self, key: str, subve: "ValidationError"):
        """
        Take an existing ValidationError and add it as a nested set of errors under the supplied key

        :param key: the key under which all the nested validation errors should go
        :param subve: the existing ValidationError object
        :return:
        """
        if key not in self._errors:
            self._errors[key] = {"errors": []}
        if "nested" not in self._errors[key]:
            self._errors[key]["nested"] = {}

        for k, v in subve.errors.items():
            self._errors[key]["nested"][k] = v

    def has_errors(self) -> bool:
        """Are there any errors registered"""
        return len(self._errors) > 0

    def __str__(self):
        return str(self._errors)


"""
Pattern to represent an ``Announce Endorsement`` notification
https://coar-notify.net/specification/1.0.0/announce-endorsement/
"""
from coarnotify.core.notify import NotifyPattern, NotifyTypes, NotifyItem, NotifyProperties, NotifyObject
from coarnotify.core.activitystreams2 import ActivityStreamsTypes, Properties
from coarnotify.exceptions import ValidationError

from typing import Union

__all__ = ["AnnounceEndorsement", "AnnounceEndorsementContext", "AnnounceEndorsementItem"]

class AnnounceEndorsement(NotifyPattern):
    """
    Class to represent an Announce Endorsement pattern
    """
    TYPE = [ActivityStreamsTypes.ANNOUNCE, NotifyTypes.ENDORSMENT_ACTION]
    """Announce Endorsement type, consisting of Activity Streams Announce and Notify Endorsement Action"""

    @property
    def context(self) -> Union["AnnounceEndorsementContext", None]:
        """
        Get a context specific to Announce Endorsement

        :return: The Announce Endorsement context object
        """
        c = self.get_property(Properties.CONTEXT)
        if c is not None:
            return AnnounceEndorsementContext(c,
                                validate_stream_on_construct=False,
                                validate_properties=self.validate_properties,
                                validators=self.validators,
                                validation_context=Properties.CONTEXT,
                                properties_by_reference=self._properties_by_reference)
        return None

    def validate(self) -> bool:
        """
        Extends the base validation to make `context` required

        :return: ``True`` if valid, otherwise raises :py:class:`coarnotify.exceptions.ValidationError`
        """
        ve = ValidationError()
        try:
            super(NotifyPattern, self).validate()
        except ValidationError as superve:
            ve = superve

        self.required_and_validate(ve, Properties.CONTEXT, self.context)

        if ve.has_errors():
            raise ve

        return True


class AnnounceEndorsementContext(NotifyObject):
    """
    Announce Endorsement context object, which extends the base NotifyObject
    to allow us to pass back a custom :py:class:`AnnounceEndorsementItem`
    """
    @property
    def item(self) -> Union["AnnounceEndorsementItem", None]:
        """
        Get a custom :py:class:`AnnounceEndorsementItem`

        :return: the Announce Endorsement Item
        """
        i = self.get_property(NotifyProperties.ITEM)
        if i is not None:
            return AnnounceEndorsementItem(i,
                              validate_stream_on_construct=False,
                              validate_properties=self.validate_properties,
                              validators=self.validators,
                              validation_context=NotifyProperties.ITEM,
                              properties_by_reference=self._properties_by_reference)
        return None


class AnnounceEndorsementItem(NotifyItem):
    """
    Announce Endorsement Item, which extends the base NotifyItem to provide
    additional validation
    """
    def validate(self) -> bool:
        """
        Extends the base validation with validation custom to Announce Endorsement notifications

        * Adds type validation, which the base NotifyItem does not apply
        * Requires the ``mediaType`` value

        :return: ``True`` if valid, otherwise raises a ValidationError
        """
        ve = ValidationError()
        try:
            super(AnnounceEndorsementItem, self).validate()
        except ValidationError as superve:
            ve = superve

        self.required_and_validate(ve, Properties.TYPE, self.type)
        self.required(ve, NotifyProperties.MEDIA_TYPE, self.media_type)

        if ve.has_errors():
            raise ve
        return True


"""
Pattern to represent an Announce Relationship notification
https://coar-notify.net/specification/1.0.0/announce-relationship/
"""
from coarnotify.core.notify import NotifyPattern, NotifyTypes, NotifyObject
from coarnotify.core.activitystreams2 import ActivityStreamsTypes, Properties
from coarnotify.exceptions import ValidationError

from typing import Union

__all__ = ["AnnounceRelationship", "AnnounceRelationshipObject"]

class AnnounceRelationship(NotifyPattern):
    """
    Class to represent an Announce Relationship notification
    """
    TYPE = [ActivityStreamsTypes.ANNOUNCE, NotifyTypes.RELATIONSHIP_ACTION]
    """Announce Relationship types, including an ActivityStreams announce and a COAR Notify Relationship Action"""

    @property
    def object(self) -> Union["AnnounceRelationshipObject", None]:
        """Custom getter to retrieve the object property as an AnnounceRelationshipObject"""
        o = self.get_property(Properties.OBJECT)
        if o is not None:
            return AnnounceRelationshipObject(o,
                                validate_stream_on_construct=False,
                                validate_properties=self.validate_properties,
                                validators=self.validators,
                                validation_context=Properties.OBJECT,
                                properties_by_reference=self._properties_by_reference)
        return None

    def validate(self) -> bool:
        """
        Extends the base validation to make `context` required

        :return: ``True`` if valid, otherwise raises :py:class:`coarnotify.exceptions.ValidationError`
        """
        ve = ValidationError()
        try:
            super(NotifyPattern, self).validate()
        except ValidationError as superve:
            ve = superve

        self.required_and_validate(ve, Properties.CONTEXT, self.context)

        if ve.has_errors():
            raise ve

        return True


class AnnounceRelationshipObject(NotifyObject):
    """
    Custom object class for Announce Relationship to apply the custom validation
    """
    def validate(self) -> bool:
        """
        Extend the base validation to include the following constraints:

        * The object triple is required and each part must validate

        :return:
        """
        ve = ValidationError()
        try:
            super(AnnounceRelationshipObject, self).validate()
        except ValidationError as superve:
            ve = superve

        self.required_and_validate(ve, Properties.TYPE, self.type)

        subject, relationship, object = self.triple
        self.required_and_validate(ve, Properties.SUBJECT_TRIPLE, subject)
        self.required_and_validate(ve, Properties.RELATIONSHIP_TRIPLE, relationship)
        self.required_and_validate(ve, Properties.OBJECT_TRIPLE, object)

        if ve.has_errors():
            raise ve

        return True



"""
Pattern to represent an Accept notification
https://coar-notify.net/specification/1.0.0/accept/
"""
from coarnotify.core.notify import NotifyPattern, NestedPatternObjectMixin
from coarnotify.core.activitystreams2 import ActivityStreamsTypes, Properties
from coarnotify.exceptions import ValidationError

__all__ = ["Accept"]

class Accept(NestedPatternObjectMixin, NotifyPattern):
    """
    Class to represent an Accept notification
    """
    TYPE = ActivityStreamsTypes.ACCEPT
    """ The Accept type """

    def validate(self) -> bool:
        """
        Validate the Accept pattern.

        In addition to the base validation, this:

        * Makes ``inReplyTo`` required
        * Requires the ``inReplyTo`` value to be the same as the ``object.id`` value

        :return: ``True`` if valid, otherwise raises a :py:class:`coarnotify.exceptions.ValidationError`
        """
        ve = ValidationError()
        try:
            super(Accept, self).validate()
        except ValidationError as superve:
            ve = superve

        # Technically, no need to validate the value, as this is handled by the superclass,
        # but leaving it in for completeness
        self.required_and_validate(ve, Properties.IN_REPLY_TO, self.in_reply_to)

        objid = self.object.id if self.object else None
        if self.in_reply_to != objid:
            ve.add_error(Properties.IN_REPLY_TO,
                         f"Expected inReplyTo id to be the same as the nested object id. inReplyTo: {self.in_reply_to}, object.id: {objid}")

        if ve.has_errors():
            raise ve

        return True


from copy import deepcopy

from coarnotify.test.fixtures import BaseFixtureFactory


class RequestEndorsementFixtureFactory(BaseFixtureFactory):
    @classmethod
    def source(cls, copy=True):
        if copy:
            return deepcopy(REQUEST_ENDORSEMENT)
        return REQUEST_ENDORSEMENT

    @classmethod
    def invalid(cls):
        source = cls.source()
        cls._base_invalid(source)
        cls._actor_invalid(source)
        cls._object_invalid(source)
        return source


REQUEST_ENDORSEMENT = {
    "@context": [
        "https://www.w3.org/ns/activitystreams",
        "https://coar-notify.net"
    ],
    "actor": {
        "id": "https://orcid.org/0000-0002-1825-0097",
        "name": "Josiah Carberry",
        "type": "Person"
    },
    "id": "urn:uuid:0370c0fb-bb78-4a9b-87f5-bed307a509dd",
    "object": {
        "id": "https://research-organisation.org/repository/preprint/201203/421/",
        "ietf:cite-as": "https://doi.org/10.5555/12345680",
        "ietf:item": {
            "id": "https://research-organisation.org/repository/preprint/201203/421/content.pdf",
            "mediaType": "application/pdf",
            "type": [
                "Article",
                "sorg:ScholarlyArticle"
            ]
        },
        "type": [
            "Page",
            "sorg:AboutPage"
        ]
    },
    "origin": {
        "id": "https://research-organisation.org/repository",
        "inbox": "https://research-organisation.org/inbox/",
        "type": "Service"
    },
    "target": {
        "id": "https://overlay-journal.com/system",
        "inbox": "https://overlay-journal.com/inbox/",
        "type": "Service"
    },
    "type": [
        "Offer",
        "coar-notify:EndorsementAction"
    ]
}



"""
HTTP layer interface and default implementation using requests lib
"""
import requests


class HttpLayer:
    """
    Interface for the HTTP layer

    This defines the methods which need to be implemented in order for the client to fully operate
    """

    def post(self, url: str, data: str, headers: dict=None, *args, **kwargs) -> 'HttpResponse':
        """
        Make an HTTP POST request to the supplied URL with the given body data, and headers

        `args` and `kwargs` can be used to pass implementation-specific parameters

        :param url: the request URL
        :param data: the body data
        :param headers: HTTP headers as a dict to include in the request
        :param args: argument list to pass on to the implementation
        :param kwargs: keyword arguments to pass on to the implementation
        :return: HttpResponse
        """
        raise NotImplementedError()

    def get(self, url: str, headers: dict=None, *args, **kwargs) -> 'HttpResponse':
        """
        Make an HTTP GET request to the supplied URL with the given headers

        `args` and `kwargs` can be used to pass implementation-specific parameters

        :param url: the request URL
        :param headers: HTTP headers as a dict to include in the request
        :param args: argument list to pass on to the implementation
        :param kwargs: keyword arguments to pass on to the implementation
        :return: HttpResponse
        """
        raise NotImplementedError()


class HttpResponse:
    """
    Interface for the HTTP response object

    This defines the methods which need to be implemented in order for the client to fully operate
    """

    def header(self, header_name: str) -> str:
        """
        Get the value of a header from the response

        :param header_name: the name of the header
        :return: the header value
        """
        raise NotImplementedError()

    @property
    def status_code(self) -> int:
        """
        Get the status code of the response

        :return: the status code
        """
        raise NotImplementedError()


#######################################
## Implementations using requests lib

class RequestsHttpLayer(HttpLayer):
    """
    Implementation of the HTTP layer using the requests library.  This is the default implementation
    used when no other implementation is supplied
    """

    def post(self, url: str, data: str, headers: dict=None, *args, **kwargs) -> 'RequestsHttpResponse':
        """
        Make an HTTP POST request to the supplied URL with the given body data, and headers

        `args` and `kwargs` can be used to pass additional parameters to the `requests.post` method,
        such as authentication credentials, etc.

        :param url: the request URL
        :param data: the body data
        :param headers: HTTP headers as a dict to include in the request
        :param args: argument list to pass on to requests.post
        :param kwargs: keyword arguments to pass on to requests.post
        :return: RequestsHttpResponse
        """
        resp = requests.post(url, data=data, headers=headers, *args, **kwargs)
        return RequestsHttpResponse(resp)

    def get(self, url: str, headers: dict=None, *args, **kwargs) -> 'RequestsHttpResponse':
        """
        Make an HTTP GET request to the supplied URL with the given headers

        `args` and `kwargs` can be used to pass additional parameters to the `requests.get` method,
        such as authentication credentials, etc.

        :param url: the request URL
        :param headers: HTTP headers as a dict to include in the request
        :param args: argument list to pass on to requests.get
        :param kwargs: keyword arguments to pass on to requests.get
        :return: HttpResponse
        """

        resp = requests.get(url, headers=headers, *args, **kwargs)
        return RequestsHttpResponse(resp)

class RequestsHttpResponse(HttpResponse):
    """
    Implementation fo the HTTP response object using the requests library

    This wraps the requests response object and provides the interface required by the client

    :param resp: response object from the requests library
    """

    def __init__(self, resp: requests.Response):
        """
        Construct the object as a wrapper around the original requests response object

        :param resp: response object from the requests library
        """
        self._resp = resp

    def header(self, header_name: str) -> str:
        """
        Get the value of a header from the response

        :param header_name: the name of the header
        :return: the header value
        """
        return self._resp.headers.get(header_name)

    @property
    def status_code(self) -> int:
        """
        Get the status code of the response

        :return: the status code
        """
        return self._resp.status_code

    @property
    def requests_response(self) -> requests.Response:
        """Get the original requests response object"""
        return self._resp



STORE_DIR = "/your/store/dir"
"""The directory on the local machine to use to store incoming JSON files"""

HOST = "localhost"
"""Host where the app will run"""

PORT = 5005
"""Port to start the app on"""

RESPONSE_STATUS = 201
"""HTTP Response to provide to any incoming reqeusts.  201 and 202 are the specification compliant values"""

VALIDATE_INCOMING = True
"""Should the server attempt to validate the incoming notifications"""

DEBUG = True
"""Put flask into debug mode for developer convenience"""

DEBUG_PYCHARM = False
"""Put the app into PyCharm debug mode.  This turns off ``DEBUG`` and starts the PyCharm debugger.  You can set this here, or you can start the test server with the ``-d`` option"""

DEBUG_PYCHARM_SERVER = "localhost"
"""The host to connect to for PyCharm debugging"""

DEBUG_PYCHARM_PORT = 6000
"""The port to connect to for PyCharm debugging"""


"""
This module is home to all the core model objects from which the notify patterns extend
"""

from coarnotify.core.activitystreams2 import ActivityStream, Properties, ActivityStreamsTypes, ACTIVITY_STREAMS_OBJECTS
from coarnotify import validate
from coarnotify.exceptions import ValidationError
from typing import Union, Tuple
import uuid
from copy import deepcopy

NOTIFY_NAMESPACE = "https://coar-notify.net"
"""Namespace for COAR Notify, to be used to construct namespaced properties used in COAR Notify Patterns"""

class NotifyProperties:
    """
    COAR Notify properties used in COAR Notify Patterns

    Most of these are provided as tuples, where the first element is the property name, and the second element is the namespace.
    Some are provided as plain strings without namespaces

    These are suitable to be used as property names in all the property getters/setters in the notify pattern objects
    and in the validation configuration.
    """
    INBOX = ("inbox", NOTIFY_NAMESPACE)
    """``inbox`` property"""

    CITE_AS = ("ietf:cite-as", NOTIFY_NAMESPACE)
    """``ietf:cite-as`` property"""

    ITEM = ("ietf:item", NOTIFY_NAMESPACE)
    """``ietf:item`` property"""

    NAME = "name"
    """``name`` property"""

    MEDIA_TYPE = "mediaType"
    """``mediaType`` property"""

class NotifyTypes:
    """
    List of all the COAR Notify types patterns may use.

    These are in addition to the base Activity Streams types, which are in :py:class:`coarnotify.core.activitystreams2.ActivityStreamsTypes`
    """
    ENDORSMENT_ACTION = "coar-notify:EndorsementAction"
    INGEST_ACTION = "coar-notify:IngestAction"
    RELATIONSHIP_ACTION = "coar-notify:RelationshipAction"
    REVIEW_ACTION = "coar-notify:ReviewAction"
    UNPROCESSABLE_NOTIFICATION = "coar-notify:UnprocessableNotification"

    ABOUT_PAGE = "sorg:AboutPage"


_VALIDATION_RULES = {
    Properties.ID: {
        "default": validate.absolute_uri,
        "context": {
            Properties.CONTEXT: {
                "default": validate.url
            },
            Properties.ORIGIN: {
                "default": validate.url
            },
            Properties.TARGET: {
                "default": validate.url
            },
            NotifyProperties.ITEM: {
                "default": validate.url
            }
        }
    },
    Properties.TYPE: {
        "default": validate.type_checker,
        "context": {
            Properties.ACTOR: {
                "default": validate.one_of([
                    ActivityStreamsTypes.SERVICE,
                    ActivityStreamsTypes.APPLICATION,
                    ActivityStreamsTypes.GROUP,
                    ActivityStreamsTypes.ORGANIZATION,
                    ActivityStreamsTypes.PERSON
                ])
            },

            Properties.OBJECT: {
                "default": validate.at_least_one_of(ACTIVITY_STREAMS_OBJECTS) #validate.contains("sorg:AboutPage"),
            },

            Properties.CONTEXT: {
                "default": validate.at_least_one_of(ACTIVITY_STREAMS_OBJECTS) #validate.contains("sorg:AboutPage"),
            },

            NotifyProperties.ITEM: {
                "default": validate.at_least_one_of(ACTIVITY_STREAMS_OBJECTS) #validate.contains("sorg:AboutPage"),
            }
        }
    },
    NotifyProperties.CITE_AS: {
        "default": validate.url
    },
    NotifyProperties.INBOX: {
        "default": validate.url
    },
    Properties.IN_REPLY_TO: {
        "default": validate.absolute_uri
    },
    Properties.SUBJECT_TRIPLE: {
        "default": validate.absolute_uri
    },
    Properties.OBJECT_TRIPLE: {
        "default": validate.absolute_uri
    },
    Properties.RELATIONSHIP_TRIPLE: {
        "default": validate.absolute_uri
    }
}

VALIDATORS: validate.Validator = validate.Validator(_VALIDATION_RULES)
"""Default Validator object for all pattern types, of type :py:class:`coarnotify.validate.Validator`"""


class NotifyBase:
    """
    Base class from which all Notify objects extend.

    There are two kinds of Notify objects:

    1. Patterns, which are the notifications themselves
    2. Pattern Parts, which are nested elements in the Patterns, such as objects, contexts, actors, etc

    This class forms the basis for both of those types, and provides essential services,
    such as construction, accessors and validation, as well as supporting the essential
    properties "id" and "type"
    """
    def __init__(self, stream: Union[ActivityStream, dict] = None,
                 validate_stream_on_construct: bool=True,
                 validate_properties: bool=True,
                 validators: validate.Validator=None,
                 validation_context: Union[str, Tuple[str, str]]=None,
                 properties_by_reference: bool=True):
        """
        Base constructor that all subclasses should call

        :param stream:  The activity stream object, or a dict from which one can be created
        :param validate_stream_on_construct:    should the incoming stream be validated at construction-time
        :param validate_properties:     should individual properties be validated as they are set
        :param validators:      the validator object for this class and all nested elements.  If not provided will use the default :py:data:`VALIDATORS`
        :param validation_context:  the context in which this object is being validated.  This is used to determine which validators to use
        :param properties_by_reference:     should properties be get and set by reference (the default) or by value.  Use this with caution: setting by value
            makes it impossible to set a property in a nested object using the dot notation, like ``obj.actor.name = "Bob"``, instead you will need to retrive
            the object, set the value, then set the whole object back on the parent object.
        """
        self._validate_stream_on_construct = validate_stream_on_construct
        self._validate_properties = validate_properties
        self._validators = validators if validators is not None else VALIDATORS
        self._validation_context = validation_context
        self._properties_by_reference = properties_by_reference
        validate_now = False

        if stream is None:
            self._stream = ActivityStream()
        elif isinstance(stream, dict):
            validate_now = validate_stream_on_construct
            self._stream = ActivityStream(stream)
        else:
            validate_now = validate_stream_on_construct
            self._stream = stream

        if self._stream.get_property(Properties.ID) is None:
            self._stream.set_property(Properties.ID, "urn:uuid:" + str(uuid.uuid4().hex))

        if validate_now:
            self.validate()

    @property
    def validate_properties(self) -> bool:
        """Are properties being validated on set"""
        return self._validate_properties

    @property
    def validate_stream_on_construct(self) -> bool:
        """Is the stream validated on construction"""
        return self._validate_stream_on_construct

    @property
    def validators(self) -> validate.Validator:
        """The validator object for this instance"""
        return self._validators

    @property
    def doc(self):
        """The underlying ActivityStream object, excluding the JSON-LD @context"""
        return self._stream.doc

    @property
    def id(self) -> str:
        """The ``id`` of the object"""
        return self.get_property(Properties.ID)

    @id.setter
    def id(self, value: str):
        self.set_property(Properties.ID, value)

    @property
    def type(self) -> Union[str, list[str]]:
        """The ``type`` of the object"""
        return self.get_property(Properties.TYPE)

    @type.setter
    def type(self, types: Union[str, list[str]]):
        self.set_property(Properties.TYPE, types)

    def get_property(self, prop_name: Union[str, Tuple[str, str]], by_reference: bool=None):
        """
        Generic property getter.  It is strongly recommended that all accessors proxy for this function
        as this enforces by-reference/by-value accessing, and mediates directly with the underlying
        activity stream object.

        :param prop_name: The property to retrieve
        :param by_reference: Whether to retrieve by_reference or by_value.  If not supplied will default to the object-wide setting
        :return: the property value
        """
        if by_reference is None:
            by_reference = self._properties_by_reference
        val = self._stream.get_property(prop_name)
        if by_reference:
            return val
        else:
            return deepcopy(val)

    def set_property(self, prop_name: Union[str, Tuple[str, str]], value, by_reference: bool=None):
        """
        Generic property setter.  It is strongly recommended that all accessors proxy for this function
        as this enforces by-reference/by-value accessing, and mediates directly with the underlying
        activity stream object.

        :param prop_name: The property to set
        :param value: The value to set
        :param by_reference: Whether to set by_reference or by_value.  If not supplied will default to the object-wide setting
        """
        if by_reference is None:
            by_reference = self._properties_by_reference
        self.validate_property(prop_name, value)
        if not by_reference:
            value = deepcopy(value)
        self._stream.set_property(prop_name, value)

    def validate(self) -> bool:
        """
        Validate the object.  This provides the basic validation on ``id`` and ``type``.
        Subclasses should override this method with their own validation, and call this method via ``super`` first to ensure
        the basic properties are validated.

        :return: ``True`` or raise a :py:class:`coarnotify.exceptions.ValidationError` if there are errors
        """
        ve = ValidationError()

        self.required_and_validate(ve, Properties.ID, self.id)
        self.required_and_validate(ve, Properties.TYPE, self.type)

        if ve.has_errors():
            raise ve
        return True

    def validate_property(self, prop_name: Union[str, Tuple[str, str]], value,
                          force_validate: bool=False, raise_error: bool=True) -> Tuple[bool, str]:
        """
        Validate a single property.  This is used internally by :py:meth:`set_property`.

        If the object has ``validate_properties`` set to ``False`` then that behaviour may be overridden by setting ``force_validate`` to ``True``

        The validator applied to the property will be determined according to the ``validators`` property of the object
        and the ``validation_context`` of the object.

        :param prop_name: The property to validate
        :param value:  the value to validate
        :param force_validate:  whether to validate anyway, even if property validation is turned off at the object level
        :param raise_error: raise an exception on validation failure, or return a tuple with the result
        :return: A tuple of whether validation was successful, and the error message if it was not
            (the empty string is returned as the second element if validation was successful)
        """
        if value is None:
            return True, ""
        if self.validate_properties or force_validate:
            validator = self.validators.get(prop_name, self._validation_context)
            if validator is not None:
                try:
                    validator(self, value)
                except ValueError as ve:
                    if raise_error:
                        raise ve
                    else:
                        return False, str(ve)
        return True, ""

    def _register_property_validation_error(self, ve: ValidationError, prop_name: Union[str, tuple], value):
        """Force validate the property and if an error is found, add it to the validation error"""
        e, msg = self.validate_property(prop_name, value, force_validate=True, raise_error=False)
        if not e:
            ve.add_error(prop_name, msg)

    def required(self, ve: ValidationError, prop_name: Union[str, tuple], value):
        """
        Add a required error to the validation error if the value is None

        :param ve: The validation error to which to add the message
        :param prop_name:   The property to check
        :param value:   The value
        """
        if value is None:
            pn = prop_name if not isinstance(prop_name, tuple) else prop_name[0]
            ve.add_error(prop_name, validate.REQUIRED_MESSAGE.format(x=pn))

    def required_and_validate(self, ve: ValidationError, prop_name: Union[str, tuple], value):
        """
        Add a required error to the validation error if the value is None, and then validate the value if not.

        Any error messages are added to the ``ValidationError`` object

        :param ve: the validation error to which to add the message
        :param prop_name: The property to check
        :param value: the value to check
        """
        if value is None:
            pn = prop_name if not isinstance(prop_name, tuple) else prop_name[0]
            ve.add_error(prop_name, validate.REQUIRED_MESSAGE.format(x=pn))
        else:
            if isinstance(value, NotifyBase):
                try:
                    value.validate()
                except ValidationError as subve:
                    ve.add_nested_errors(prop_name, subve)
            else:
                self._register_property_validation_error(ve, prop_name, value)

    def optional_and_validate(self, ve: ValidationError, prop_name: Union[str, tuple], value):
        """
        Validate the value if it is not None, but do not raise a validation error if it is None

        :param ve:
        :param prop_name:
        :param value:
        :return:
        """
        if value is not None:
            if isinstance(value, NotifyBase):
                try:
                    value.validate()
                except ValidationError as subve:
                    ve.add_nested_errors(prop_name, subve)
            else:
                self._register_property_validation_error(ve, prop_name, value)

    def to_jsonld(self) -> dict:
        """
        Get the notification pattern as JSON-LD

        :return: JSON-LD representation of the pattern
        """
        return self._stream.to_jsonld()


class NotifyPattern(NotifyBase):
    """
    Base class for all notification patterns
    """
    TYPE = ActivityStreamsTypes.OBJECT
    """The type of the pattern.  This should be overridden by subclasses, otherwise defaults to ``Object``"""

    def __init__(self, stream: Union[ActivityStream, dict] = None,
                 validate_stream_on_construct=True,
                 validate_properties=True,
                 validators=None,
                 validation_context=None,
                 properties_by_reference=True):
        """
        Constructor for the NotifyPattern

        This constructor will ensure that the pattern has its mandated type :py:attr:`TYPE` in the ``type`` property

        :param stream:  The activity stream object, or a dict from which one can be created
        :param validate_stream_on_construct:    should the incoming stream be validated at construction-time
        :param validate_properties:     should individual properties be validated as they are set
        :param validators:      the validator object for this class and all nested elements.  If not provided will use the default :py:data:`VALIDATORS`
        :param validation_context:  the context in which this object is being validated.  This is used to determine which validators to use
        :param properties_by_reference:     should properties be get and set by reference (the default) or by value.  Use this with caution: setting by value
            makes it impossible to set a property in a nested object using the dot notation, like ``obj.actor.name = "Bob"``, instead you will need to retrive
            the object, set the value, then set the whole object back on the parent object.
        """
        super(NotifyPattern, self).__init__(stream=stream,
                                            validate_stream_on_construct=validate_stream_on_construct,
                                            validate_properties=validate_properties,
                                            validators=validators,
                                            validation_context=validation_context,
                                            properties_by_reference=properties_by_reference)
        self._ensure_type_contains(self.TYPE)

    def _ensure_type_contains(self, types: Union[str, list[str]]):
        """Ensure that the type field contains the given types"""
        existing = self._stream.get_property(Properties.TYPE)
        if existing is None:
            self.set_property(Properties.TYPE, types)
        else:
            if not isinstance(existing, list):
                existing = [existing]
            if not isinstance(types, list):
                types = [types]
            for t in types:
                if t not in existing:
                    existing.append(t)
            if len(existing) == 1:
                existing = existing[0]
            self.set_property(Properties.TYPE, existing)

    @property
    def origin(self) -> Union["NotifyService", None]:
        """Get the origin property of the notification"""
        o = self.get_property(Properties.ORIGIN)
        if o is not None:
            return NotifyService(o,
                                 validate_stream_on_construct=False,
                                 validate_properties=self.validate_properties,
                                 validators=self.validators,
                                 validation_context=Properties.ORIGIN,
                                 properties_by_reference=self._properties_by_reference)
        return None

    @origin.setter
    def origin(self, value: "NotifyService"):
        self.set_property(Properties.ORIGIN, value.doc)

    @property
    def target(self) -> Union["NotifyService", None]:
        """Get the target property of the notification"""
        t = self.get_property(Properties.TARGET)
        if t is not None:
            return NotifyService(t,
                                 validate_stream_on_construct=False,
                                 validate_properties=self.validate_properties,
                                 validators=self.validators,
                                 validation_context=Properties.TARGET,
                                 properties_by_reference=self._properties_by_reference)
        return None

    @target.setter
    def target(self, value: "NotifyService"):
        self.set_property(Properties.TARGET, value.doc)

    @property
    def object(self) -> Union["NotifyObject", None]:
        """Get the object property of the notification"""
        o = self.get_property(Properties.OBJECT)
        if o is not None:
            return NotifyObject(o,
                            validate_stream_on_construct=False,
                            validate_properties=self.validate_properties,
                            validators=self.validators,
                            validation_context=Properties.OBJECT,
                            properties_by_reference=self._properties_by_reference)
        return None

    @object.setter
    def object(self, value: "NotifyObject"):
        self.set_property(Properties.OBJECT, value.doc)

    @property
    def in_reply_to(self) -> str:
        """Get the inReplyTo property of the notification"""
        return self.get_property(Properties.IN_REPLY_TO)

    @in_reply_to.setter
    def in_reply_to(self, value: str):
        self.set_property(Properties.IN_REPLY_TO, value)

    @property
    def actor(self) -> Union["NotifyActor", None]:
        """Get the actor property of the notification"""
        a = self.get_property(Properties.ACTOR)
        if a is not None:
            return NotifyActor(a,
                            validate_stream_on_construct=False,
                            validate_properties=self.validate_properties,
                            validators=self.validators,
                            validation_context=Properties.ACTOR,
                            properties_by_reference=self._properties_by_reference)
        return None

    @actor.setter
    def actor(self, value: "NotifyActor"):
        self.set_property(Properties.ACTOR, value.doc)

    @property
    def context(self) -> Union["NotifyObject", None]:
        """Get the context property of the notification"""
        c = self.get_property(Properties.CONTEXT)
        if c is not None:
            return NotifyObject(c,
                            validate_stream_on_construct=False,
                            validate_properties=self.validate_properties,
                            validators=self.validators,
                            validation_context=Properties.CONTEXT,
                            properties_by_reference=self._properties_by_reference)
        return None

    @context.setter
    def context(self, value: "NotifyObject"):
        self.set_property(Properties.CONTEXT, value.doc)

    def validate(self) -> bool:
        """
        Base validator for all notification patterns.  This extends the validate function on the superclass.

        In addition to the base class's constraints, this applies the following validation:

        * The ``origin``, ``target`` and ``object`` properties are required and must be valid
        * The ``actor`` ``inReplyTo`` and ``context`` properties are optional, but if present must be valid

        :py:class:`NotifyBase`
        :return: ``True`` if valid, otherwise raises :py:class:`coarnotify.exceptions.ValidationError`
        """
        ve = ValidationError()
        try:
            super(NotifyPattern, self).validate()
        except ValidationError as superve:
            ve = superve

        self.required_and_validate(ve, Properties.ORIGIN, self.origin)
        self.required_and_validate(ve, Properties.TARGET, self.target)
        self.required_and_validate(ve, Properties.OBJECT, self.object)
        self.optional_and_validate(ve, Properties.ACTOR, self.actor)
        self.optional_and_validate(ve, Properties.IN_REPLY_TO, self.in_reply_to)
        self.optional_and_validate(ve, Properties.CONTEXT, self.context)

        if ve.has_errors():
            raise ve

        return True

class NotifyPatternPart(NotifyBase):
    """
    Base class for all pattern parts, such as objects, contexts, actors, etc

    If there is a default type specified, and a type is not given at construction, then
    the default type will be added

    :param stream:  The activity stream object, or a dict from which one can be created
    :param validate_stream_on_construct:    should the incoming stream be validated at construction-time
    :param validate_properties:     should individual properties be validated as they are set
    :param validators:      the validator object for this class and all nested elements.  If not provided will use the default :py:data:`VALIDATORS`
    :param validation_context:  the context in which this object is being validated.  This is used to determine which validators to use
    :param properties_by_reference:     should properties be get and set by reference (the default) or by value.  Use this with caution: setting by value
        makes it impossible to set a property in a nested object using the dot notation, like ``obj.actor.name = "Bob"``, instead you will need to retrive
        the object, set the value, then set the whole object back on the parent object.
    """
    DEFAULT_TYPE = None
    """The default type for this object, if none is provided on construction.  If not provided, then no default type will be set"""

    ALLOWED_TYPES = []
    """The list of types that are permissable for this object.  If the list is empty, then any type is allowed"""

    def __init__(self, stream: Union[ActivityStream, dict] = None,
                 validate_stream_on_construct=True,
                 validate_properties=True,
                 validators=None,
                 validation_context=None,
                 properties_by_reference=True):
        """
        Constructor for the NotifyPatternPart

        If there is a default type specified, and a type is not given at construction, then
        the default type will be added

        :param stream:  The activity stream object, or a dict from which one can be created
        :param validate_stream_on_construct:    should the incoming stream be validated at construction-time
        :param validate_properties:     should individual properties be validated as they are set
        :param validators:      the validator object for this class and all nested elements.  If not provided will use the default :py:data:`VALIDATORS`
        :param validation_context:  the context in which this object is being validated.  This is used to determine which validators to use
        :param properties_by_reference:     should properties be get and set by reference (the default) or by value.  Use this with caution: setting by value
            makes it impossible to set a property in a nested object using the dot notation, like ``obj.actor.name = "Bob"``, instead you will need to retrive
            the object, set the value, then set the whole object back on the parent object.
        """
        super(NotifyPatternPart, self).__init__(stream=stream,
                                                validate_stream_on_construct=validate_stream_on_construct,
                                                validate_properties=validate_properties,
                                                validators=validators,
                                                validation_context=validation_context,
                                                properties_by_reference=properties_by_reference)
        if self.DEFAULT_TYPE is not None and self.type is None:
            self.type = self.DEFAULT_TYPE

    @NotifyBase.type.setter
    def type(self, types: Union[str, list[str]]):
        """Set the type of the object, and validate that it is one of the allowed types if present"""
        if not isinstance(types, list):
            types = [types]

        if len(self.ALLOWED_TYPES) > 0:
            for t in types:
                if t not in self.ALLOWED_TYPES:
                    raise ValueError(f"Type value {t} is not one of the permitted values")

        # keep single values as single values, not lists
        if len(types) == 1:
            types = types[0]

        self.set_property(Properties.TYPE, types)


class NotifyService(NotifyPatternPart):
    """
    Default class to represent a service in the COAR Notify pattern.

    Services are used to represent ``origin`` and ``target`` properties in the notification patterns

    Specific patterns may need to extend this class to provide their specific behaviours and validation
    """
    DEFAULT_TYPE = ActivityStreamsTypes.SERVICE
    """The default type for a service is ``Service``, but the type can be set to any value"""

    @property
    def inbox(self) -> str:
        """Get the ``inbox`` property of the service"""
        return self.get_property(NotifyProperties.INBOX)

    @inbox.setter
    def inbox(self, value: str):
        self.set_property(NotifyProperties.INBOX, value)


class NotifyObject(NotifyPatternPart):
    """
    Deafult class to represent an object in the COAR Notify pattern.  Objects can be used for ``object`` or ``context`` properties
    in notify patterns

    Specific patterns may need to extend this class to provide their specific behaviours and validation
    """

    @property
    def cite_as(self) -> str:
        """Get the ``ietf:cite-as`` property of the object"""
        return self.get_property(NotifyProperties.CITE_AS)

    @cite_as.setter
    def cite_as(self, value: str):
        self.set_property(NotifyProperties.CITE_AS, value)

    @property
    def item(self) -> Union["NotifyItem", None]:
        """Get the ``ietf:item`` property of the object"""
        i = self.get_property(NotifyProperties.ITEM)
        if i is not None:
            return NotifyItem(i,
                            validate_stream_on_construct=False,
                            validate_properties=self.validate_properties,
                            validators=self.validators,
                            validation_context=NotifyProperties.ITEM,
                            properties_by_reference=self._properties_by_reference)
        return None

    @item.setter
    def item(self, value: "NotifyItem"):
        self.set_property(NotifyProperties.ITEM, value)

    @property
    def triple(self) -> tuple[str, str, str]:
        """Get object, relationship and subject properties as a relationship triple"""
        obj = self.get_property(Properties.OBJECT_TRIPLE)
        rel = self.get_property(Properties.RELATIONSHIP_TRIPLE)
        subj = self.get_property(Properties.SUBJECT_TRIPLE)
        return obj, rel, subj

    @triple.setter
    def triple(self, value: tuple[str, str, str]):
        obj, rel, subj = value
        self.set_property(Properties.OBJECT_TRIPLE, obj)
        self.set_property(Properties.RELATIONSHIP_TRIPLE, rel)
        self.set_property(Properties.SUBJECT_TRIPLE, subj)

    def validate(self) -> bool:
        """
        Validate the object.  This overrides the base validation, as objects only absolutely require an ``id`` property,
        so the base requirement for a ``type`` is relaxed.

        :return: ``True`` if valid, otherwise raises :py:class:`coarnotify.exceptions.ValidationError`
        """
        ve = ValidationError()

        self.required_and_validate(ve, Properties.ID, self.id)

        if ve.has_errors():
            raise ve
        return True


class NotifyActor(NotifyPatternPart):
    """
    Deafult class to represents an actor in the COAR Notify pattern.
    Actors are used to represent the ``actor`` property in the notification patterns

    Specific patterns may need to extend this class to provide their specific behaviours and validation
    """
    DEFAULT_TYPE = ActivityStreamsTypes.SERVICE
    """Default type is ``Service``, but can also be set as any one of the other allowed types"""

    ALLOWED_TYPES = [DEFAULT_TYPE,
                     ActivityStreamsTypes.APPLICATION,
                     ActivityStreamsTypes.GROUP,
                     ActivityStreamsTypes.ORGANIZATION,
                     ActivityStreamsTypes.PERSON
                     ]
    """The allowed types for an actor: ``Service``, ``Application``, ``Group``, ``Organisation``, ``Person``"""

    @property
    def name(self) -> str:
        """Get the name property of the actor"""
        return self.get_property(NotifyProperties.NAME)

    @name.setter
    def name(self, value: str):
        self.set_property(NotifyProperties.NAME, value)


class NotifyItem(NotifyPatternPart):
    """
    Defult class to represent an item in the COAR Notify pattern.
    Items are used to represent the ``ietf:item`` property in the notification patterns

    Specific patterns may need to extend this class to provide their specific behaviours and validation
    """
    @property
    def media_type(self) -> str:
        """Get the ``mediaType`` property of the item"""
        return self.get_property(NotifyProperties.MEDIA_TYPE)

    @media_type.setter
    def media_type(self, value: str):
        self.set_property(NotifyProperties.MEDIA_TYPE, value)

    def validate(self):
        """
        Validate the item.  This overrides the base validation, as objects only absolutely require an ``id`` property,
        so the base requirement for a ``type`` is relaxed.

        :return: ``True`` if valid, otherwise raises :py:class:`coarnotify.exceptions.ValidationError`
        """
        ve = ValidationError()

        self.required_and_validate(ve, Properties.ID, self.id)

        if ve.has_errors():
            raise ve
        return True


## Mixins
##########################################################

class NestedPatternObjectMixin:
    """
    A mixin to add to a pattern which can override the default object property to return a full
    nested pattern from the ``object`` property, rather than the default :py:class:`NotifyObject`

    This mixin needs to be first on the inheritance list, as it overrides the object property
    of the NotifyPattern class.

    For example:

    .. code-block:: python

        class MySpecialPattern(NestedPatternObjectMixin, NotifyPattern):
            pass
    """
    @property
    def object(self) -> Union[NotifyPattern, NotifyObject, None]:
        """Retrieve an object as it's correctly typed pattern, falling back to a default ``NotifyObject`` if no pattern matches"""
        o = self.get_property(Properties.OBJECT)
        if o is not None:
            from coarnotify.factory import COARNotifyFactory  # late import to avoid circular dependency
            nested = COARNotifyFactory.get_by_object(deepcopy(o),
                                                     validate_stream_on_construct=False,
                                                     validate_properties=self.validate_properties,
                                                     validators=self.validators,
                                                     validation_context=None)  # don't supply a validation context, as these objects are not typical nested objects
            if nested is not None:
                return nested

            # if we are unable to construct the typed nested object, just return a generic object
            return NotifyObject(deepcopy(o),
                                validate_stream_on_construct=False,
                                validate_properties=self.validate_properties,
                                validators=self.validators,
                                validation_context=Properties.OBJECT)
        return None

    @object.setter
    def object(self, value: Union[NotifyObject, NotifyPattern]):
        self.set_property(Properties.OBJECT, value.doc)


class SummaryMixin:
    """
    Mixin to provide an API for setting and getting the ``summary`` property of a pattern
    """
    @property
    def summary(self) -> str:
        """The summary property of the pattern"""
        return self.get_property(Properties.SUMMARY)

    @summary.setter
    def summary(self, summary: str):
        self.set_property(Properties.SUMMARY, summary)


"""
Factory for producing the correct model based on the type or data within a payload
"""

from typing import List, Callable, Union
from coarnotify.core.activitystreams2 import ActivityStream, Properties
from coarnotify.core.notify import NotifyPattern
from coarnotify.patterns import (
    Accept,
    AnnounceEndorsement,
    AnnounceRelationship,
    AnnounceReview,
    AnnounceServiceResult,
    Reject,
    RequestEndorsement,
    RequestReview,
    TentativelyAccept,
    TentativelyReject,
    UnprocessableNotification,
    UndoOffer
)
from coarnotify.exceptions import NotifyException


class COARNotifyFactory:
    """
    Factory for producing the correct model based on the type or data within a payload
    """

    MODELS = [
        Accept,
        AnnounceEndorsement,
        AnnounceRelationship,
        AnnounceReview,
        AnnounceServiceResult,
        Reject,
        RequestEndorsement,
        RequestReview,
        TentativelyAccept,
        TentativelyReject,
        UnprocessableNotification,
        UndoOffer
    ]
    """The list of model classes recognised by this factory"""

    @classmethod
    def get_by_types(cls, incoming_types:Union[str, List[str]]) -> Union[Callable, None]:
        """
        Get the model class based on the supplied types.  The returned callable is the class, not an instance.

        This is achieved by inspecting all of the known types in ``MODELS``, and performing the following
        calculation:

        1. If the supplied types are a subset of the model types, then this is a candidate, keep a reference to it
        2. If the candidate fit is exact (supplied types and model types are the same), return the class
        3. If the class is a better fit than the last candidate, update the candidate.  If the fit is exact, return the class
        4. Once we have run out of models to check, return the best candidate (or None if none found)

        :param incoming_types: a single type or list of types.  If a list is provided, ALL types must match a candidate
        :return:    A class representing the best fit for the supplied types, or ``None`` if no match
        """
        if not isinstance(incoming_types, list):
            incoming_types = [incoming_types]

        candidate = None
        candidate_fit = None

        for m in cls.MODELS:
            document_types = m.TYPE
            if not isinstance(document_types, list):
                document_types = [document_types]
            if set(document_types).issubset(set(incoming_types)):
                if candidate_fit is None:
                    candidate = m
                    candidate_fit = len(incoming_types) - len(document_types)
                    if candidate_fit == 0:
                        return candidate

                else:
                    fit = len(incoming_types) - len(document_types)
                    if fit == 0:
                        return m
                    if abs(fit) < abs(candidate_fit):
                        candidate = m
                        candidate_fit = fit

        return candidate

    @classmethod
    def get_by_object(cls, data: dict, *args, **kwargs) -> NotifyPattern:
        """
        Get an instance of a model based on the data provided.

        Internally this calls ``get_by_types`` to determine the class to instantiate, and then creates an instance of that
        Using the supplied args and kwargs.

        If a model cannot be found that matches the data, a NotifyException is raised.

        :param data: The raw stream data to parse and instantiate around
        :param args: any args to pass to the object constructor
        :param kwargs: any kwargs to pass to the object constructor
        :return: A NotifyPattern of the correct type, wrapping the data
        """
        stream = ActivityStream(data)

        types = stream.get_property(Properties.TYPE)
        if types is None:
            raise NotifyException("No type found in object")

        klazz = cls.get_by_types(types)

        inst = klazz(data, *args, **kwargs)
        return inst

    @classmethod
    def register(cls, model: NotifyPattern):
        existing = cls.get_by_types(model.TYPE)
        if existing is not None:
            cls.MODELS.remove(existing)
        cls.MODELS.append(model)


"""
Pattern to represent the Announce Service Result notification
https://coar-notify.net/specification/1.0.0/announce-resource/
"""
from coarnotify.core.notify import NotifyPattern, NotifyItem, NotifyProperties, NotifyObject
from coarnotify.core.activitystreams2 import ActivityStreamsTypes, Properties
from coarnotify.exceptions import ValidationError

from typing import Union

__all__ = ["AnnounceServiceResult", "AnnounceServiceResultObject", "AnnounceServiceResultContext", "AnnounceServiceResultItem"]

class AnnounceServiceResult(NotifyPattern):
    """
    Class to represent the Announce Service Result Pattern
    """

    TYPE = ActivityStreamsTypes.ANNOUNCE
    """Announce Service Result type, the ActivityStreams Announce type"""

    @property
    def object(self) -> Union["AnnounceServiceResultObject", None]:
        """
        Custom getter to retrieve the object property as an AnnounceServiceResultObject

        :return: AnnounceServiceResultObject
        """
        o = self.get_property(Properties.OBJECT)
        if o is not None:
            return AnnounceServiceResultObject(o,
                                        validate_stream_on_construct=False,
                                        validate_properties=self.validate_properties,
                                        validators=self.validators,
                                        validation_context=Properties.OBJECT,
                                        properties_by_reference=self._properties_by_reference)
        return None

    @property
    def context(self) -> Union["AnnounceServiceResultContext", None]:
        """
        Custom getter to retrieve the context property as an AnnounceServiceResultContext

        :return:    AnnounceSericeResultCOntext
        """
        c = self.get_property(Properties.CONTEXT)
        if c is not None:
            return AnnounceServiceResultContext(c,
                                         validate_stream_on_construct=False,
                                         validate_properties=self.validate_properties,
                                         validators=self.validators,
                                         validation_context=Properties.CONTEXT,
                                         properties_by_reference=self._properties_by_reference)
        return None

    def validate(self) -> bool:
        """
        Extends the base validation to make `context` required

        :return: ``True`` if valid, otherwise raises :py:class:`coarnotify.exceptions.ValidationError`
        """
        ve = ValidationError()
        try:
            super(NotifyPattern, self).validate()
        except ValidationError as superve:
            ve = superve

        self.required_and_validate(ve, Properties.CONTEXT, self.context)

        if ve.has_errors():
            raise ve

        return True


class AnnounceServiceResultContext(NotifyObject):
    """
    Custom object class for Announce Service Result to provide the custom item getter
    """

    @property
    def item(self) -> Union["AnnounceServiceResultItem", None]:
        """
        Custom getter to retrieve the item property as an AnnounceServiceResultItem
        :return:
        """
        i = self.get_property(NotifyProperties.ITEM)
        if i is not None:
            return AnnounceServiceResultItem(i,
                              validate_stream_on_construct=False,
                              validate_properties=self.validate_properties,
                              validators=self.validators,
                              validation_context=NotifyProperties.ITEM,
                              properties_by_reference=self._properties_by_reference)
        return None

class AnnounceServiceResultItem(NotifyItem):
    """
    Custom item class for Announce Service Result to apply the custom validation
    """

    def validate(self) -> bool:
        """
        Beyond the base validation, apply the following:

        * Make type required and avlid
        * Make the ``mediaType`` required

        :return: ``True`` if validation passes, else raise a ``ValidationError``
        """
        ve = ValidationError()
        try:
            super(AnnounceServiceResultItem, self).validate()
        except ValidationError as superve:
            ve = superve

        self.required_and_validate(ve, Properties.TYPE, self.type)
        self.required(ve, NotifyProperties.MEDIA_TYPE, self.media_type)

        if ve.has_errors():
            raise ve
        return True

class AnnounceServiceResultObject(NotifyObject):
    """
    Custom object class for Announce Service Result to apply the custom validation
    """

    def validate(self) -> bool:
        """
        Extend the base validation to include the following constraints:

        * The object type is required and must validate

        :return: ``True`` if validation passes, else raise a ``ValidationError``
        """
        ve = ValidationError()
        try:
            super(AnnounceServiceResultObject, self).validate()
        except ValidationError as superve:
            ve = superve

        self.required_and_validate(ve, Properties.TYPE, self.type)

        if ve.has_errors():
            raise ve

        return True


"""
This module contains everything COAR Notify needs to know about ActivityStreams 2.0
https://www.w3.org/TR/activitystreams-core/

It provides knowledge of the essential AS properties and types, and a class to wrap
ActivityStreams objects and provide a simple interface to work with them.

**NOTE** this is not a complete implementation of AS 2.0, it is **only** what is required
to work with COAR Notify patterns.
"""
from typing import Union


ACTIVITY_STREAMS_NAMESPACE = "https://www.w3.org/ns/activitystreams"
"""Namespace for Actvitity Streams, to be used to construct namespaced properties used in COAR Notify Patterns"""

class Properties:
    """
    ActivityStreams 2.0 properties used in COAR Notify Patterns

    These are provided as tuples, where the first element is the property name, and the second element is the namespace.

    These are suitbale to be used as property names in all the property getters/setters in the notify pattern objects
    and in the validation configuration.
    """
    ID = ("id", ACTIVITY_STREAMS_NAMESPACE)
    """``id`` property"""

    TYPE = ("type", ACTIVITY_STREAMS_NAMESPACE)
    """``type`` property"""

    ORIGIN = ("origin", ACTIVITY_STREAMS_NAMESPACE)
    """``origin`` property"""

    OBJECT = ("object", ACTIVITY_STREAMS_NAMESPACE)
    """``object`` property"""

    TARGET = ("target", ACTIVITY_STREAMS_NAMESPACE)
    """``target`` property"""

    ACTOR = ("actor", ACTIVITY_STREAMS_NAMESPACE)
    """``actor`` property"""

    IN_REPLY_TO = ("inReplyTo", ACTIVITY_STREAMS_NAMESPACE)
    """``inReplyTo`` property"""

    CONTEXT = ("context", ACTIVITY_STREAMS_NAMESPACE)
    """``context`` property"""

    SUMMARY = ("summary", ACTIVITY_STREAMS_NAMESPACE)
    """``summary`` property"""

    SUBJECT_TRIPLE = ("as:subject", ACTIVITY_STREAMS_NAMESPACE)
    """``as:subject`` property"""

    OBJECT_TRIPLE = ("as:object", ACTIVITY_STREAMS_NAMESPACE)
    """``as:object`` property"""

    RELATIONSHIP_TRIPLE = ("as:relationship", ACTIVITY_STREAMS_NAMESPACE)
    """``as:relationship`` property"""

class ActivityStreamsTypes:
    """
    List of all the Activity Streams types COAR Notify may use.

    Note that COAR Notify also has its own custom types and they are defined in
    :py:class:`coarnotify.models.notify.NotifyTypes`
    """

    # Activities
    ACCEPT = "Accept"
    ANNOUNCE = "Announce"
    REJECT = "Reject"
    OFFER = "Offer"
    TENTATIVE_ACCEPT = "TentativeAccept"
    TENTATIVE_REJECT = "TentativeReject"
    FLAG = "Flag"
    UNDO = "Undo"

    # Objects
    ACTIVITY = "Activity"
    APPLICATION = "Application"
    ARTICLE = "Article"
    AUDIO = "Audio"
    COLLECTION = "Collection"
    COLLECTION_PAGE = "CollectionPage"
    RELATIONSHIP = "Relationship"
    DOCUMENT = "Document"
    EVENT = "Event"
    GROUP = "Group"
    IMAGE = "Image"
    INTRANSITIVE_ACTIVITY = "IntransitiveActivity"
    NOTE = "Note"
    OBJECT = "Object"
    ORDERED_COLLECTION = "OrderedCollection"
    ORDERED_COLLECTION_PAGE = "OrderedCollectionPage"
    ORGANIZATION = "Organization"
    PAGE = "Page"
    PERSON = "Person"
    PLACE = "Place"
    PROFILE = "Profile"
    QUESTION = "Question"
    SERVICE = "Service"
    TOMBSTONE = "Tombstone"
    VIDEO = "Video"

ACTIVITY_STREAMS_OBJECTS = [
    ActivityStreamsTypes.ACTIVITY,
    ActivityStreamsTypes.APPLICATION,
    ActivityStreamsTypes.ARTICLE,
    ActivityStreamsTypes.AUDIO,
    ActivityStreamsTypes.COLLECTION,
    ActivityStreamsTypes.COLLECTION_PAGE,
    ActivityStreamsTypes.RELATIONSHIP,
    ActivityStreamsTypes.DOCUMENT,
    ActivityStreamsTypes.EVENT,
    ActivityStreamsTypes.GROUP,
    ActivityStreamsTypes.IMAGE,
    ActivityStreamsTypes.INTRANSITIVE_ACTIVITY,
    ActivityStreamsTypes.NOTE,
    ActivityStreamsTypes.OBJECT,
    ActivityStreamsTypes.ORDERED_COLLECTION,
    ActivityStreamsTypes.ORDERED_COLLECTION_PAGE,
    ActivityStreamsTypes.ORGANIZATION,
    ActivityStreamsTypes.PAGE,
    ActivityStreamsTypes.PERSON,
    ActivityStreamsTypes.PLACE,
    ActivityStreamsTypes.PROFILE,
    ActivityStreamsTypes.QUESTION,
    ActivityStreamsTypes.SERVICE,
    ActivityStreamsTypes.TOMBSTONE,
    ActivityStreamsTypes.VIDEO
]
"""The sub-list of ActivityStreams types that are also objects in AS 2.0"""

class ActivityStream:
    """
    A simple wrapper around an ActivityStreams dictionary object

    Construct it with a python dictionary that represents an ActivityStreams object, or
    without to create a fresh, blank object.

    :param raw: the raw ActivityStreams object, as a dictionary
    """
    def __init__(self, raw: dict=None):
        """
        Construct a new ActivityStream object

        :param raw: the raw ActivityStreams object, as a dictionary
        """
        self._doc = raw if raw is not None else {}
        self._context = []
        if "@context" in self._doc:
            self._context = self._doc["@context"]
            if not isinstance(self._context, list):
                self._context = [self._context]
            del self._doc["@context"]

    @property
    def doc(self) -> dict:
        """The internal dictionary representation of the ActivityStream, without the json-ld context"""
        return self._doc

    @doc.setter
    def doc(self, doc:dict):
        self._doc = doc

    @property
    def context(self):
        """The json-ld context of the ActivityStream"""
        return self._context

    @context.setter
    def context(self, context):
        self._context = context

    def _register_namespace(self, namespace: Union[str, tuple[str, str]]):
        """
        Register a namespace in the context of the ActivityStream
        """
        entry = namespace
        if isinstance(namespace, tuple):
            url = namespace[1]
            short = namespace[0]
            entry = {short: url}

        if entry not in self._context:
            self._context.append(entry)

    def set_property(self, property: Union[str, tuple[str, str], tuple[str, tuple[str, str]]], value):
        """
        Set an arbitrary property on the object.  The property name can be one of:

        * A simple string with the property name
        * A tuple of the property name and the full namespace ``("name", "http://example.com/ns")``
        * A tuple containing the property name and another tuple of the short name and the full namespace ``("name", ("as", "http://example.com/ns"))``

        :param property: the property name
        :param value: the value to set
        """
        prop_name = property
        namespace = None
        if isinstance(property, tuple):
            prop_name = property[0]
            namespace = property[1]

        self._doc[prop_name] = value
        if namespace is not None:
            self._register_namespace(namespace)

    def get_property(self, property: Union[str, tuple[str, str], tuple[str, tuple[str, str]]]):
        """
        Get an arbitrary property on the object.  The property name can be one of:

        * A simple string with the property name
        * A tuple of the property name and the full namespace ``("name", "http://example.com/ns")``
        * A tuple containing the property name and another tuple of the short name and the full namespace ``("name", ("as", "http://example.com/ns"))``

        :param property:   the property name
        :return: the value of the property, or None if it does not exist
        """
        prop_name = property
        namespace = None
        if isinstance(property, tuple):
            prop_name = property[0]
            namespace = property[1]

        return self._doc.get(prop_name, None)

    def to_jsonld(self) -> dict:
        """
        Get the activity stream as a JSON-LD object

        :return:
        """
        return {
            "@context": self._context,
            **self._doc
        }


"""
Pattern to represent the Undo Offer notification
https://coar-notify.net/specification/1.0.0/undo-offer/
"""
from coarnotify.core.notify import NotifyPattern, NestedPatternObjectMixin, SummaryMixin
from coarnotify.core.activitystreams2 import ActivityStreamsTypes, Properties
from coarnotify.exceptions import ValidationError

__all__ = ["UndoOffer"]

class UndoOffer(NestedPatternObjectMixin, NotifyPattern, SummaryMixin):
    """
    Class to represent the Undo Offer notification
    """
    TYPE = ActivityStreamsTypes.UNDO
    """Undo Offer type, the ActivityStreams Undo type"""

    def validate(self) -> bool:
        """
        In addition to the base validation apply the following constraints:

        * The ``inReplyTo`` property is required
        * The ``inReplyTo`` value must match the ``object.id`` value

        :return:
        """
        ve = ValidationError()
        try:
            super(UndoOffer, self).validate()
        except ValidationError as superve:
            ve = superve

        # Technically, no need to validate the value, as this is handled by the superclass,
        # but leaving it in for completeness
        self.required_and_validate(ve, Properties.IN_REPLY_TO, self.in_reply_to)

        objid = self.object.id if self.object else None
        if self.in_reply_to != objid:
            ve.add_error(Properties.IN_REPLY_TO, f"Expected inReplyTo id to be the same as the nested object id. inReplyTo: {self.in_reply_to}, object.id: {objid}")

        if ve.has_errors():
            raise ve

        return True



from setuptools import setup, find_packages

setup(
    name="coarnotify",
    version="1.0.1.3",  # Version 3 of the library for the 1.0.1 spec
    packages=find_packages(),
    install_requires=[
        "requests"
    ],
    urls=["https://coar-notify.net/", "http://cottagelabs.com/"],
    author="Cottage Labs",
    author_email="richard@cottagelabs.com",
    description="COAR Notify Common Library",
    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    license="Apache2",
    classifiers=[],
    extras_require={
        'docs': [
            'sphinx',
            'sphinx-autoapi'
        ],
        'test': [
            "Flask>3.0.0"
        ],
    }
)



"""
All the COAR Notify pattern objects are defined in this module.

Some of the pattern objects have supporting objects in their individual submodules
"""
from coarnotify.patterns.accept import Accept
from coarnotify.patterns.announce_endorsement import AnnounceEndorsement
from coarnotify.patterns.announce_relationship import AnnounceRelationship
from coarnotify.patterns.announce_review import AnnounceReview
from coarnotify.patterns.announce_service_result import AnnounceServiceResult
from coarnotify.patterns.reject import Reject
from coarnotify.patterns.request_endorsement import RequestEndorsement
from coarnotify.patterns.request_review import RequestReview
from coarnotify.patterns.tentatively_accept import TentativelyAccept
from coarnotify.patterns.tentatively_reject import TentativelyReject
from coarnotify.patterns.unprocessable_notification import UnprocessableNotification
from coarnotify.patterns.undo_offer import UndoOffer



from unittest import TestCase

from coarnotify.core.notify import NotifyPattern
from coarnotify.patterns import (
    Accept,
    AnnounceEndorsement,
    AnnounceRelationship,
    AnnounceReview,
    AnnounceServiceResult,
    Reject,
    RequestEndorsement,
    RequestReview,
    TentativelyAccept,
    TentativelyReject,
    UnprocessableNotification,
    UndoOffer
)
from coarnotify.factory import COARNotifyFactory

from coarnotify.test.fixtures import (
    AcceptFixtureFactory,
    AnnounceEndorsementFixtureFactory,
    AnnounceRelationshipFixtureFactory,
    AnnounceReviewFixtureFactory,
    AnnounceServiceResultFixtureFactory,
    RejectFixtureFactory,
    RequestEndorsementFixtureFactory,
    RequestReviewFixtureFactory,
    TentativelyAcceptFixtureFactory,
    TentativelyRejectFixtureFactory,
    UnprocessableNotificationFixtureFactory,
    UndoOfferFixtureFactory
)


class TestFactory(TestCase):
    def test_01_accept(self):
        acc = COARNotifyFactory.get_by_types(Accept.TYPE)
        assert acc == Accept

        source = AcceptFixtureFactory.source()
        acc = COARNotifyFactory.get_by_object(source)
        assert isinstance(acc, Accept)

        assert acc.id == source["id"]

    def test_02_announce_endorsement(self):
        ae = COARNotifyFactory.get_by_types(AnnounceEndorsement.TYPE)
        assert ae == AnnounceEndorsement

        source = AnnounceEndorsementFixtureFactory.source()
        ae = COARNotifyFactory.get_by_object(source)
        assert isinstance(ae, AnnounceEndorsement)

        assert ae.id == source["id"]

    def test_04_announce_relationship(self):
        ar = COARNotifyFactory.get_by_types(AnnounceRelationship.TYPE)
        assert ar == AnnounceRelationship

        source = AnnounceRelationshipFixtureFactory.source()
        ar = COARNotifyFactory.get_by_object(source)
        assert isinstance(ar, AnnounceRelationship)

        assert ar.id == source["id"]

    def test_05_announce_review(self):
        ar = COARNotifyFactory.get_by_types(AnnounceReview.TYPE)
        assert ar == AnnounceReview

        source = AnnounceReviewFixtureFactory.source()
        ar = COARNotifyFactory.get_by_object(source)
        assert isinstance(ar, AnnounceReview)

        assert ar.id == source["id"]

    def test_06_announce_service_result(self):
        ar = COARNotifyFactory.get_by_types(AnnounceServiceResult.TYPE)
        assert ar == AnnounceServiceResult

        source = AnnounceServiceResultFixtureFactory.source()
        ar = COARNotifyFactory.get_by_object(source)
        assert isinstance(ar, AnnounceServiceResult)

        assert ar.id == source["id"]

    def test_07_reject(self):
        ar = COARNotifyFactory.get_by_types(Reject.TYPE)
        assert ar == Reject

        source = RejectFixtureFactory.source()
        ar = COARNotifyFactory.get_by_object(source)
        assert isinstance(ar, Reject)

        assert ar.id == source["id"]

    def test_08_request_endorsement(self):
        ar = COARNotifyFactory.get_by_types(RequestEndorsement.TYPE)
        assert ar == RequestEndorsement

        source = RequestEndorsementFixtureFactory.source()
        ar = COARNotifyFactory.get_by_object(source)
        assert isinstance(ar, RequestEndorsement)

        assert ar.id == source["id"]

    def test_10_request_review(self):
        ar = COARNotifyFactory.get_by_types(RequestReview.TYPE)
        assert ar == RequestReview

        source = RequestReviewFixtureFactory.source()
        ar = COARNotifyFactory.get_by_object(source)
        assert isinstance(ar, RequestReview)

        assert ar.id == source["id"]

    def test_11_tentatively_accept(self):
        ar = COARNotifyFactory.get_by_types(TentativelyAccept.TYPE)
        assert ar == TentativelyAccept

        source = TentativelyAcceptFixtureFactory.source()
        ar = COARNotifyFactory.get_by_object(source)
        assert isinstance(ar, TentativelyAccept)

        assert ar.id == source["id"]

    def test_12_tentatively_reject(self):
        ar = COARNotifyFactory.get_by_types(TentativelyReject.TYPE)
        assert ar == TentativelyReject

        source = TentativelyRejectFixtureFactory.source()
        ar = COARNotifyFactory.get_by_object(source)
        assert isinstance(ar, TentativelyReject)

        assert ar.id == source["id"]

    def test_13_unprocessable_notification(self):
        ar = COARNotifyFactory.get_by_types(UnprocessableNotification.TYPE)
        assert ar == UnprocessableNotification

        source = UnprocessableNotificationFixtureFactory.source()
        ar = COARNotifyFactory.get_by_object(source)
        assert isinstance(ar, UnprocessableNotification)

        assert ar.id == source["id"]

    def test_14_undo_offer(self):
        ar = COARNotifyFactory.get_by_types(UndoOffer.TYPE)
        assert ar == UndoOffer

        source = UndoOfferFixtureFactory.source()
        ar = COARNotifyFactory.get_by_object(source)
        assert isinstance(ar, UndoOffer)

        assert ar.id == source["id"]

    def test_15_register(self):
        class TestPattern(NotifyPattern):
            TYPE = Accept.TYPE

        COARNotifyFactory.register(TestPattern)

        tp = COARNotifyFactory.get_by_types(Accept.TYPE)
        assert tp == TestPattern




from coarnotify.http import HttpLayer, HttpResponse


class MockHttpLayer(HttpLayer):
    def __init__(self, status_code=200, location=None):
        self._status_code = status_code
        self._location = location

    def post(self, url, data, headers=None, *args, **kwargs):
        return MockHttpResponse(status_code=self._status_code, location=self._location)

    def get(self, url, headers=None, *args, **kwargs):
        raise NotImplementedError()

    def head(self, url, headers=None, *args, **kwargs):
        raise NotImplementedError()


class MockHttpResponse(HttpResponse):
    def __init__(self, status_code=200, location=None):
        self._status_code = status_code
        self._location = location

    def header(self, header_name):
        if header_name.lower() == "location":
            return self._location

    @property
    def status_code(self):
        return self._status_code


from copy import deepcopy
from coarnotify.core.notify import NotifyObject, NotifyService


class NotifyFixtureFactory:
    @classmethod
    def source(cls):
        return deepcopy(BASE_NOTIFY)

    @classmethod
    def target(cls):
        return NotifyService(deepcopy(BASE_NOTIFY["target"]))

    @classmethod
    def origin(cls):
        return NotifyService(deepcopy(BASE_NOTIFY["origin"]))

    @classmethod
    def object(cls):
        return NotifyObject(deepcopy(BASE_NOTIFY["object"]))


BASE_NOTIFY = {
    "@context": [
        "https://www.w3.org/ns/activitystreams",
        "https://purl.org/coar/notify"
    ],
    "id": "urn:uuid:94ecae35-dcfd-4182-8550-22c7164fe23f",
    "type": "Object",
    "origin": {
        "id": "https://overlay-journal.com/system",
        "inbox": "https://overlay-journal.com/inbox/",
        "type": "Service"
    },
    "object": {
        "id": "https://overlay-journal.com/articles/00001/",
        "ietf:cite-as": "https://overlay-journal.com/articles/00001/",
        "type": [
            "Page",
            "sorg:WebPage"
        ]
    },
    "target": {
        "id": "https://research-organisation.org/repository",
        "inbox": "https://research-organisation.org/inbox/",
        "type": "Service"
    },
    "actor": {
        "id": "https://overlay-journal.com",
        "name": "Overlay Journal",
        "type": "Service"
    },
    "inReplyTo": "urn:uuid:0370c0fb-bb78-4a9b-87f5-bed307a509dd",
    "context": {
        "id": "https://research-organisation.org/repository/preprint/201203/421/",
        "ietf:cite-as": "https://doi.org/10.5555/12345680",
        "ietf:item": {
            "id": "https://research-organisation.org/repository/preprint/201203/421/content.pdf",
            "mediaType": "application/pdf",
            "type": [
                "Article",
                "sorg:ScholarlyArticle"
            ]
        },
        "type": "sorg:AboutPage"
    }
}